"""
The file `datahandler.py` is an abstract class for Datahandler.

Author : Thivin Anandh D

Date : 03/May/2024

History : 03/May/2024 - Initial implementation with basic data handling
"""

from abc import abstractmethod


class DataHandler:
    """
    This class is to handle data for 2D problems, convert them into tensors using custom tf functions.
    It is responsible for all type conversions and data handling.

    .. note:: All inputs to these functions are generally numpy arrays with dtype np.float64.
              So we can either maintain the same dtype or convert them to tf.float32 ( for faster computation ).

    :param fespace: The FESpace2D object.
    :type fespace: FESpace2D
    :param domain: The Domain2D object.
    :type domain: Domain2D
    :param shape_val_mat_list: List of shape function values for each cell.
    :type shape_val_mat_list: list
    :param grad_x_mat_list: List of shape function derivatives with respect to x for each cell.
    :type grad_x_mat_list: list
    :param grad_y_mat_list: List of shape function derivatives with respect to y for each cell.
    :type grad_y_mat_list: list
    :param x_pde_list: List of actual coordinates of the quadrature points for each cell.
    :type x_pde_list: list
    :param forcing_function_list: List of forcing function values for each cell.
    :type forcing_function_list: list
    :param dtype: The tensorflow dtype to be used for all the tensors.
    :type dtype: tf.DType
    """

    def __init__(self, fespace, domain, dtype):
        """
        Constructor for the DataHandler class

        :param fespace: The FESpace2D object.
        :type fespace: FESpace2D
        :param domain: The Domain object.
        :type domain: Domain2D
        :param dtype: The tensorflow dtype to be used for all the tensors.
        :type dtype: tf.DType
        """

        self.fespace = fespace
        self.domain = domain
        self.dtype = dtype

    @abstractmethod
    def get_dirichlet_input(self):
        """
        This function will return the input for the Dirichlet boundary data

        :return:
            - input_dirichlet (tf.Tensor): The input for the Dirichlet boundary data
            - actual_dirichlet (tf.Tensor): The actual Dirichlet boundary data
        """

    @abstractmethod
    def get_test_points(self):
        """
        Get the test points for the given domain.

        :return: The test points for the given domain.
        :rtype: tf.Tensor
        """

    @abstractmethod
    def get_bilinear_params_dict_as_tensors(self, function):
        """
        Accepts a function from example file and converts all the values into tensors of the given dtype

        :param function: The function from the example file which returns the bilinear parameters dictionary
        :type function: function
        :return: The bilinear parameters dictionary with all the values converted to tensors
        :rtype: dict
        """

    @abstractmethod
    def get_sensor_data(self, exact_sol, num_sensor_points, mesh_type, file_name=None):
        """
        Accepts a function from example file and converts all the values into tensors of the given dtype

        :param exact_sol: The function from the example file which returns the exact solution
        :type exact_sol: function
        :param num_sensor_points: The number of sensor points to be generated
        :type num_sensor_points: int
        :param mesh_type: The type of mesh to be used for sensor data generation
        :type mesh_type: str
        :param file_name: The name of the file to be used for external mesh generation, defaults to None
        :type file_name: str, optional
        :return: The sensor points and sensor values as tensors
        :rtype: tuple[tf.Tensor, tf.Tensor]
        """

    @abstractmethod
    def get_inverse_params(self, inverse_params_dict_function):
        """
        Accepts a function from example file and converts all the values into tensors of the given dtype

        :param inverse_params_dict_function: The function from the example file which returns the inverse parameters dictionary
        :type inverse_params_dict_function: function
        :return: The inverse parameters dictionary with all the values converted to tensors
        :rtype: dict
        """



"""
The file `quadratureformulas.py` contains class for all quadrature formulas

Author: Thivin Anandh D

Changelog: Not specified

Known issues: None

Dependencies: numpy, scipy
"""

from abc import abstractmethod


class Quadratureformulas:
    """
    Defines the Quadrature Formulas for the 2D Quadrilateral elements.

    :param quad_order: The order of the quadrature.
    :type quad_order: int
    :param quad_type: The type of the quadrature.
    :type quad_type: str
    """

    def __init__(self, quad_order: int, quad_type: str, num_quad_points: int):
        """
        Constructor for the Quadratureformulas_Quad2D class.

        :param quad_order: The order of the quadrature.
        :type quad_order: int
        :param quad_type: The type of the quadrature.
        :type quad_type: str
        """
        self.quad_order = quad_order
        self.quad_type = quad_type
        self.num_quad_points = num_quad_points

    @abstractmethod
    def get_quad_values(self):
        """
        Returns the quadrature weights, xi and eta values.

        :return: A tuple containing the quadrature weights, xi values, and eta values.
        :rtype: tuple
        """

    @abstractmethod
    def get_num_quad_points(self):
        """
        Returns the number of quadrature points.

        :return: The number of quadrature points.
        :rtype: int
        """



"""
This module `fe2d_setup_main.py` will be used to setup the FE2D and quadrature rule for a given cell based on the
given mesh and the degree of the basis functions.

Author: Thivin Anandh D

Date: 30/Aug/2023

"""

# Importing the required libraries
from .basis_function_2d import *

# import Quadrature rules
from .quadratureformulas_quad2d import *


# import base class for FE transformation
from .fe_transformation_2d import *


class FE2DSetupMain:
    """
    This class is used to setup the FE2D and quadrature rule for a given cell based on the given mesh and the degree of the basis functions.
    """

    def __init__(
        self, cell_type: str, fe_order: int, fe_type: str, quad_order: int, quad_type: str
    ):
        self.cell_type = cell_type
        self.fe_order = fe_order
        self.fe_type = fe_type
        self.quad_order = quad_order
        self.quad_type = quad_type

        self.assign_basis_function()

    def assign_basis_function(self) -> BasisFunction2D:
        """
        Assigns the basis function based on the cell type and the fe_order.

        :return: An instance of the BasisFunction2D class representing the assigned basis function.
        :rtype: BasisFunction2D
        :raises ValueError: If the fe_order is invalid.
        """
        # check for FE order lower bound and higher bound
        if self.fe_order <= 1 or self.fe_order >= 1e3:
            print(f"Invalid FE order {self.fe_order} in {self.__class__.__name__} from {__name__}.")
            raise ValueError("FE order should be greater than 1 and less than 1e4.")

        if self.cell_type == "quadrilateral":
            self.n_nodes = 4

            # --- LEGENDRE --- #
            if self.fe_type == "legendre" or self.fe_type == "jacobi":
                # jacobi is added for backward compatibility with prev pushes
                # generally, jacobi is referred to as Legendre basis on previous iterations
                return Basis2DQNLegendre(self.fe_order**2)

            elif self.fe_type == "legendre_special":
                return Basis2DQNLegendreSpecial(self.fe_order**2)

            # ----- CHEBYSHEV ---- #
            elif self.fe_type == "chebyshev_2":
                return Basis2DQNChebyshev2(self.fe_order**2)

            # ----- PLain jacobi ---- #
            elif self.fe_type == "jacobi_plain":
                return Basis2DQNJacobi(self.fe_order**2)

            else:
                print(
                    f"Invalid FE order {self.fe_order} in {self.__class__.__name__} from {__name__}."
                )
                raise ValueError(
                    'FE order should be one of the : "legendre" , "jacobi", "legendre_special", "chebyshev_2", "jacobi_plain"'
                )

        print(f"Invalid cell type {self.cell_type} in {self.__class__.__name__} from {__name__}.")

    def assign_quadrature_rules(self):
        """
        Assigns the quadrature rule based on the quad_order.

        :return: A tuple containing the weights, xi, and eta values of the quadrature rule.
        :rtype: tuple
        :raises ValueError: If the quad_order is invalid or the cell_type is invalid.
        """
        if self.cell_type == "quadrilateral":
            if self.quad_order < 3:
                raise ValueError("Quad order should be greater than 2.")
            elif self.quad_order >= 2 and self.quad_order <= 9999:
                weights, xi, eta = Quadratureformulas_Quad2D(
                    self.quad_order, self.quad_type
                ).get_quad_values()
                return weights, xi, eta
            else:
                print(
                    f"Invalid quad order {self.quad_order} in {self.__class__.__name__} from {__name__}."
                )
                raise ValueError("Quad order should be between 1 and 9999.")

        raise ValueError(
            f"Invalid cell type {self.cell_type} in {self.__class__.__name__} from {__name__}."
        )

    def assign_fe_transformation(
        self, fe_transformation_type, cell_coordinates
    ) -> FETransforamtion2D:
        """
        Assigns the FE transformation based on the cell type.

        :param fe_transformation_type: The type of FE transformation.
        :type fe_transformation_type: str
        :param cell_coordinates: The coordinates of the cell.
        :type cell_coordinates: list
        :return: The FE transformation object.
        :rtype: FETransforamtion2D
        :raises ValueError: If the cell type or FE transformation type is invalid.
        """
        if self.cell_type == "quadrilateral":
            if fe_transformation_type == "affine":
                return QuadAffin(cell_coordinates)
            elif fe_transformation_type == "bilinear":
                return QuadBilinear(cell_coordinates)
            else:
                raise ValueError(
                    f"Invalid FE transformation type {fe_transformation_type} in {self.__class__.__name__} from {__name__}."
                )

        else:
            raise ValueError(
                f"Invalid cell type {self.cell_type} in {self.__class__.__name__} from {__name__}."
            )



'''
This module provides an interface for the transformation methods used in the
2D finite element analysis. The transformation is essential for mapping
element geometry in the reference domain to the actual physical domain.
The primary functionalities encapsulated within this class include:
    1. set_cell() - To set the physical coordinates of the cell. These coordinates are essential for subsequent computations involving Jacobians and transformations.
    2. get_original_from_ref(xi, eta) - Given reference coordinates (xi, eta), this method returns the corresponding coordinates in the physical domain.
    3. get_jacobian(xi, eta) - For a given point in the reference domain, represented by (xi, eta), this method calculates and returns the Jacobian of the transformation, which provides information about the local stretching, rotation, and skewing of the element. Further implementations of this class for specific element types (like quad and triangular elements) can incorporate more detailed and element-specific transformation techniques.

Author: Thivin Anandh D

Date:   20/Sep/2023

History: First version -  20/Sep/2023 - Thivin Anandh
'''

from abc import abstractmethod


class FETransforamtion2D:
    """
    This class represents a 2D finite element transformation.
    """

    def __init__(self):
        """
        Constructor for the FETransforamtion2D class.
        """

    @abstractmethod
    def set_cell(self):
        """
        Set the cell coordinates, which will be used to calculate the Jacobian and actual values.

        :return: None
        """

    @abstractmethod
    def get_original_from_ref(self, xi, eta):
        """
        This method returns the original coordinates from the reference coordinates.

        :param xi: The xi coordinate in the reference space.
        :type xi: float
        :param eta: The eta coordinate in the reference space.
        :type eta: float
        :return: The original coordinates (x, y) corresponding to the given reference coordinates.
        :rtype: tuple
        """

    @abstractmethod
    def get_original_from_ref(self, xi, eta):
        """
        This method returns the original coordinates from the reference coordinates.

        :param xi: The xi value of the reference coordinates.
        :type xi: float
        :param eta: The eta value of the reference coordinates.
        :type eta: float
        :return: The original coordinates corresponding to the given reference coordinates.
        :rtype: tuple
        """

    @abstractmethod
    def get_jacobian(self, xi, eta):
        """
        This method returns the Jacobian of the transformation.

        :param xi: The xi coordinate.
        :type xi: float
        :param eta: The eta coordinate.
        :type eta: float
        :return: The Jacobian matrix.
        :rtype: numpy.ndarray
        """


## Mandatory, Import all the basis functions here (Quad element Transformations)
from .quad_affine import *
from .quad_bilinear import *



"""
This file `poisson2d.py` is implementation of our efficient tensor-based loss calculation for poisson equation

Author: Thivin Anandh D

Date: 21/Sep/2023

History: Initial implementation

Refer: https://arxiv.org/abs/2404.12063
"""

import tensorflow as tf


# PDE loss function for the poisson problem
@tf.function
def pde_loss_poisson(
    test_shape_val_mat,
    test_grad_x_mat,
    test_grad_y_mat,
    pred_nn,
    pred_grad_x_nn,
    pred_grad_y_nn,
    forcing_function,
    bilinear_params,
):  # pragma: no cover
    """
    This method returns the loss for the Poisson Problem of the PDE
    """
    # ∫du/dx. dv/dx dΩ
    pde_diffusion_x = tf.transpose(tf.linalg.matvec(test_grad_x_mat, pred_grad_x_nn))

    # ∫du/dy. dv/dy dΩ
    pde_diffusion_y = tf.transpose(tf.linalg.matvec(test_grad_y_mat, pred_grad_y_nn))

    # eps * ∫ (du/dx. dv/dx + du/dy. dv/dy) dΩ
    pde_diffusion = bilinear_params["eps"] * (pde_diffusion_x + pde_diffusion_y)

    residual_matrix = pde_diffusion - forcing_function

    residual_cells = tf.reduce_mean(tf.square(residual_matrix), axis=0)

    return residual_cells



"""
The file `basis_2d_QN_Jacobi.py` contains the class Basis2DQNJacobi which is used
to define the basis functions for a Jacobi Polynomial. Test functions and derivatives
are inferred from the work by Ehsan Kharazmi et.al (hp-VPINNs: Variational
Physics-Informed Neural Networks With Domain Decomposition) available at
https://github.com/ehsankharazmi/hp-VPINNs/

Author: Thivin Anandh D

Changelog: 30/Aug/2023 - Initial version

Known issues: None

Dependencies: scipy, numpy
"""

# import the jacobi polynomials
from scipy.special import jacobi

import numpy as np
from .basis_function_2d import BasisFunction2D


class Basis2DQNJacobi(BasisFunction2D):
    """
    This class defines the basis functions for a 2D QN element.
    """

    def __init__(self, num_shape_functions: int):
        super().__init__(num_shape_functions)

    def jacobi_wrapper(self, n, a, b, x):
        """
        Evaluate the Jacobi polynomial of degree n with parameters a and b at the given points x.

        :param n: Degree of the Jacobi polynomial.
        :type n: int
        :param a: First parameter of the Jacobi polynomial.
        :type a: float
        :param b: Second parameter of the Jacobi polynomial.
        :type b: float
        :param x: Points at which to evaluate the Jacobi polynomial.
        :type x: array_like

        :return: Values of the Jacobi polynomial at the given points x.
        :rtype: array_like
        """
        x = np.array(x, dtype=np.float64)
        return jacobi(n, a, b)(x)

    # Derivative of the Jacobi polynomials
    def djacobi(self, n, a, b, x, k: int):
        """
        Evaluate the k-th derivative of the Jacobi polynomial of degree n with parameters a and b at the given points x.

        :param n: Degree of the Jacobi polynomial.
        :type n: int
        :param a: First parameter of the Jacobi polynomial.
        :type a: float
        :param b: Second parameter of the Jacobi polynomial.
        :type b: float
        :param x: Points at which to evaluate the Jacobi polynomial.
        :type x: array_like
        :param k: Order of the derivative.
        :type k: int

        :return: Values of the k-th derivative of the Jacobi polynomial at the given points x.
        :rtype: array_like

        :raises ValueError: If the derivative order is not 1 or 2.

        :raises ImportError: If the required module 'jacobi' is not found.

        :raises Exception: If an unknown error occurs during the computation.
        """
        x = np.array(x, dtype=np.float64)
        if k == 1:
            return jacobi(n, a, b).deriv()(x)
        if k == 2:
            return jacobi(n, a, b).deriv(2)(x)
        else:
            print(f"Invalid derivative order {k} in {__name__}.")
            raise ValueError("Derivative order should be 1 or 2.")

    ## Helper Function
    def test_fcnx(self, n_test, x):
        """
        Compute the x-component of the test functions for a given number of test functions and x-coordinates.

        :param n_test: Number of test functions.
        :type n_test: int
        :param x: x-coordinates at which to evaluate the test functions.
        :type x: array_like

        :return: Values of the x-component of the test functions.
        :rtype: array_like
        """
        test_total = []
        for n in range(1, n_test + 1):
            test = self.jacobi_wrapper(n - 1, 0, 0, x)
            test_total.append(test)
        return np.asarray(test_total, np.float64)

    def test_fcny(self, n_test, y):
        """
        Compute the y-component of the test functions for a given number of test functions and y-coordinates.

        Parameters:
            n_test (int): Number of test functions.
            y (array_like): y-coordinates at which to evaluate the test functions.

        Returns:
            array_like: Values of the y-component of the test functions.
        """
        test_total = []
        for n in range(1, n_test + 1):
            test = self.jacobi_wrapper(n - 1, 0, 0, y)
            test_total.append(test)
        return np.asarray(test_total, np.float64)

    def dtest_fcn(self, n_test, x):
        """
        Compute the x-derivatives of the test functions for a given number of test functions and x-coordinates.

        :param n_test: Number of test functions.
        :type n_test: int
        :param x: x-coordinates at which to evaluate the test functions.
        :type x: array_like

        :return: Values of the x-derivatives of the test functions.
        :rtype: array_like
        """
        d1test_total = []
        for n in range(1, n_test + 1):
            d1test = self.djacobi(n - 1, 0, 0, x, 1)
            d1test_total.append(d1test)
        return np.asarray(d1test_total)

    def ddtest_fcn(self, n_test, x):
        """
        Compute the x-derivatives of the test functions for a given number of test functions and x-coordinates.

        :param n_test: Number of test functions.
        :type n_test: int
        :param x: x-coordinates at which to evaluate the test functions.
        :type x: array_like

        :return: Values of the x-derivatives of the test functions.
        :rtype: array_like
        """
        d1test_total = []
        for n in range(1, n_test + 1):
            d1test = self.djacobi(n - 1, 0, 0, x, 2)
            d1test_total.append(d1test)
        return np.asarray(d1test_total)

    def value(self, xi, eta):
        """
        This method returns the values of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like
        :param eta: y-coordinates at which to evaluate the basis functions.
        :type eta: array_like
        :return: Values of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        test_x = self.test_fcnx(num_shape_func_in_1d, xi)
        test_y = self.test_fcny(num_shape_func_in_1d, eta)
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                test_x[i, :] * test_y
            )

        return values

    def gradx(self, xi, eta):
        """
        This method returns the x-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like
        :param eta: y-coordinates at which to evaluate the basis functions.
        :type eta: array_like
        :return: Values of the x-derivatives of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        grad_test_x = self.dtest_fcn(num_shape_func_in_1d, xi)
        test_y = self.test_fcny(num_shape_func_in_1d, eta)
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                grad_test_x[i, :] * test_y
            )

        return values

    def grady(self, xi, eta):
        """
        This method returns the y-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like

        :param eta: y-coordinates at which to evaluate the basis functions.
        :type eta: array_like

        :return: Values of the y-derivatives of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        test_x = self.test_fcnx(num_shape_func_in_1d, xi)
        grad_test_y = self.dtest_fcn(num_shape_func_in_1d, eta)
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                test_x[i, :] * grad_test_y
            )

        return values

    def gradxx(self, xi, eta):
        """
        This method returns the xx-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like
        :param eta: y-coordinates at which to evaluate the basis functions.
        :type eta: array_like

        :return: Values of the xx-derivatives of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        grad_grad_x = self.ddtest_fcn(num_shape_func_in_1d, xi)
        test_y = self.test_fcny(num_shape_func_in_1d, eta)
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                grad_grad_x[i, :] * test_y
            )

        return values

    def gradxy(self, xi, eta):
        """
        This method returns the xy-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like
        :param eta: y-coordinates at which to evaluate the basis functions.
        :type eta: array_like
        :return: Values of the xy-derivatives of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        grad_test_x = self.dtest_fcn(num_shape_func_in_1d, xi)
        grad_test_y = self.dtest_fcn(num_shape_func_in_1d, eta)
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                grad_test_x[i, :] * grad_test_y
            )

        return values

    def gradyy(self, xi, eta):
        """
        This method returns the yy-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like
        :param eta: y-coordinates at which to evaluate the basis functions.
        :type eta: array_like

        :return: Values of the yy-derivatives of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        test_x = self.test_fcnx(num_shape_func_in_1d, xi)
        grad_grad_y = self.ddtest_fcn(num_shape_func_in_1d, eta)
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                test_x[i, :] * grad_grad_y
            )

        return values



"""
file: basis_function_3d.py
description: Abstract class for all basis functions 3D
authors: Thivin Anandh D
changelog: 03/May/2024 - First version
known_issues: None
dependencies: None specified.
"""

from abc import abstractmethod


class BasisFunction3D:  # pragma: no cover
    """
    Represents a basis function in 3D.

    Args:
        num_shape_functions (int): The number of shape functions.

    Methods:
        value(xi, eta, zeta): Evaluates the basis function at the given xi and eta coordinates.
        gradx(xi, eta, zeta): Computes the partial derivative of the basis function with respect to xi.
        grady(xi, eta, zeta): Computes the partial derivative of the basis function with respect to eta.
        gradxx(xi, eta, zeta): Computes the second partial derivative of the basis function with respect to xi.
        gradxy(xi, eta, zeta): Computes the mixed partial derivative of the basis function with respect to xi and eta.
        gradyy(xi, eta, zeta): Computes the second partial derivative of the basis function with respect to eta.
    """

    def __init__(self, num_shape_functions):
        self.num_shape_functions = num_shape_functions

    @abstractmethod
    def value(self, xi, eta, zeta):
        """
        Evaluates the basis function at the given xi and eta coordinates.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The value of the basis function at the given coordinates.
        :rtype: float
        """

    @abstractmethod
    def gradx(self, xi, eta, zeta):
        """
        Computes the partial derivative of the basis function with respect to xi.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The partial derivative of the basis function with respect to xi.
        :rtype: float
        """

    @abstractmethod
    def grady(self, xi, eta, zeta):
        """
        Computes the partial derivative of the basis function with respect to eta.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The partial derivative of the basis function with respect to eta.
        :rtype: float
        """

    @abstractmethod
    def gradxx(self, xi, eta, zeta):
        """
        Computes the second partial derivative of the basis function with respect to xi.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The second partial derivative of the basis function with respect to xi.
        :rtype: float
        """

    @abstractmethod
    def gradxy(self, xi, eta, zeta):
        """
        Computes the mixed partial derivative of the basis function with respect to xi and eta.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The mixed partial derivative of the basis function with respect to xi and eta.
        :rtype: float
        """

    @abstractmethod
    def gradyy(self, xi, eta, zeta):
        """
        Computes the second partial derivative of the basis function with respect to eta.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The second partial derivative of the basis function with respect to eta.
        :rtype: float
        """



"""
The file `model_inverse_domain.py` hosts the Neural Network (NN) model and the training loop for variational Physics-Informed Neural Networks (PINNs).
This focuses on training variational PINNs for inverse problems where the inverse parameter is constant over the entire domain.
The focus is on the model architecture and the training loop, and not on the loss functions.

Author: Thivin Anandh D

Changelog: 22/Sep/2023 - Initial implementation with basic model architecture and training loop

Known issues: None

Dependencies: None specified"""

import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras import initializers
import copy


# Custom Model
class DenseModel_Inverse_Domain(tf.keras.Model):
    """
    A subclass of tf.keras.Model that defines a dense model for an inverse problem.

    :param list layer_dims: The dimensions of the layers in the model.
    :param dict learning_rate_dict: A dictionary containing the learning rates.
    :param dict params_dict: A dictionary containing the parameters of the model.
    :param function loss_function: The loss function to be used in the model.
    :param list input_tensors_list: A list of input tensors.
    :param list orig_factor_matrices: The original factor matrices.
    :param list force_function_list: A list of force functions.
    :param list sensor_list: A list of sensors for the inverse problem.
    :param dict inverse_params_dict: A dictionary containing the parameters for the inverse problem.
    :param tf.DType tensor_dtype: The data type of the tensors.
    :param bool use_attention: Whether to use attention mechanism in the model. Defaults to False.
    :param str activation: The activation function to be used in the model. Defaults to 'tanh'.
    :param bool hessian: Whether to use Hessian in the model. Defaults to False.
    """

    def __init__(
        self,
        layer_dims,
        learning_rate_dict,
        params_dict,
        loss_function,
        input_tensors_list,
        orig_factor_matrices,
        force_function_list,
        sensor_list,  # for inverse problem
        tensor_dtype,
        use_attention=False,
        activation="tanh",
        hessian=False,
    ):
        super(DenseModel_Inverse_Domain, self).__init__()
        self.layer_dims = layer_dims
        self.use_attention = use_attention
        self.activation = activation
        self.layer_list = []
        self.loss_function = loss_function
        self.hessian = hessian

        self.tensor_dtype = tensor_dtype

        self.sensor_list = sensor_list
        # obtain sensor values
        self.sensor_points = sensor_list[0]
        self.sensor_values = sensor_list[1]

        # if dtype is not a valid tensorflow dtype, raise an error
        if not isinstance(self.tensor_dtype, tf.DType):
            raise TypeError("The given dtype is not a valid tensorflow dtype")

        self.orig_factor_matrices = orig_factor_matrices
        self.shape_function_mat_list = copy.deepcopy(orig_factor_matrices[0])
        self.shape_function_grad_x_factor_mat_list = copy.deepcopy(orig_factor_matrices[1])
        self.shape_function_grad_y_factor_mat_list = copy.deepcopy(orig_factor_matrices[2])

        self.force_function_list = force_function_list

        self.input_tensors_list = input_tensors_list
        self.input_tensor = copy.deepcopy(input_tensors_list[0])
        self.dirichlet_input = copy.deepcopy(input_tensors_list[1])
        self.dirichlet_actual = copy.deepcopy(input_tensors_list[2])

        self.params_dict = params_dict

        self.pre_multiplier_val = self.shape_function_mat_list
        self.pre_multiplier_grad_x = self.shape_function_grad_x_factor_mat_list
        self.pre_multiplier_grad_y = self.shape_function_grad_y_factor_mat_list

        self.force_matrix = self.force_function_list

        print(f"{'-'*74}")
        print(f"| {'PARAMETER':<25} | {'SHAPE':<25} |")
        print(f"{'-'*74}")
        print(
            f"| {'input_tensor':<25} | {str(self.input_tensor.shape):<25} | {self.input_tensor.dtype}"
        )
        print(
            f"| {'force_matrix':<25} | {str(self.force_matrix.shape):<25} | {self.force_matrix.dtype}"
        )
        print(
            f"| {'pre_multiplier_grad_x':<25} | {str(self.pre_multiplier_grad_x.shape):<25} | {self.pre_multiplier_grad_x.dtype}"
        )
        print(
            f"| {'pre_multiplier_grad_y':<25} | {str(self.pre_multiplier_grad_y.shape):<25} | {self.pre_multiplier_grad_y.dtype}"
        )
        print(
            f"| {'pre_multiplier_val':<25} | {str(self.pre_multiplier_val.shape):<25} | {self.pre_multiplier_val.dtype}"
        )
        print(
            f"| {'dirichlet_input':<25} | {str(self.dirichlet_input.shape):<25} | {self.dirichlet_input.dtype}"
        )
        print(
            f"| {'dirichlet_actual':<25} | {str(self.dirichlet_actual.shape):<25} | {self.dirichlet_actual.dtype}"
        )
        print(f"{'-'*74}")

        self.n_cells = params_dict["n_cells"]

        ## ----------------------------------------------------------------- ##
        ## ---------- LEARNING RATE AND OPTIMISER FOR THE MODEL ------------ ##
        ## ----------------------------------------------------------------- ##

        # parse the learning rate dictionary
        self.learning_rate_dict = learning_rate_dict
        initial_learning_rate = learning_rate_dict["initial_learning_rate"]
        use_lr_scheduler = learning_rate_dict["use_lr_scheduler"]
        decay_steps = learning_rate_dict["decay_steps"]
        decay_rate = learning_rate_dict["decay_rate"]
        # staircase = learning_rate_dict["staircase"]

        if use_lr_scheduler:
            learning_rate_fn = tf.keras.optimizers.schedules.ExponentialDecay(
                initial_learning_rate, decay_steps, decay_rate, staircase=True
            )
        else:
            learning_rate_fn = initial_learning_rate

        self.optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate_fn)

        # build the model using the input shape of the first layer in self.layer_dims
        input_shape = (None, self.layer_dims[0])
        # build the model
        self.build(input_shape=input_shape)
        # Compile the model
        self.compile(optimizer=self.optimizer)
        # print model summary
        self.summary()

        ## ----------------------------------------------------------------- ##
        ## --------------------- MODEL ARCHITECTURE ------------------------ ##
        ## ----------------------------------------------------------------- ##

        # Build dense layers based on the input list
        for dim in range(len(self.layer_dims) - 2):
            self.layer_list.append(
                layers.Dense(
                    self.layer_dims[dim + 1],
                    activation=self.activation,
                    kernel_initializer="glorot_uniform",
                    dtype=self.tensor_dtype,
                    bias_initializer="zeros",
                )
            )

        # Add a output layer with no activation
        self.layer_list.append(
            layers.Dense(
                self.layer_dims[-1],
                activation=None,
                kernel_initializer="glorot_uniform",
                dtype=self.tensor_dtype,
                bias_initializer="zeros",
            )
        )

        # Add attention layer if required
        if self.use_attention:
            self.attention_layer = layers.Attention()

        # Compile the model
        self.compile(optimizer=self.optimizer)
        self.build(input_shape=(None, self.layer_dims[0]))

        # print the summary of the model
        self.summary()

    def call(self, inputs):
        """
        The call method for the model.

        :param inputs: The input tensor for the model.
        :type inputs: tf.Tensor
        :return: The output tensor of the model.
        :rtype: tf.Tensor
        """
        x = inputs

        # Apply attention layer after input if flag is True
        if self.use_attention:
            x = self.attention_layer([x, x])

        # Loop through the dense layers
        for layer in self.layer_list:
            x = layer(x)

        return x

    def get_config(self):
        """
        Get the configuration of the model.

        Returns:
            dict: The configuration of the model.
        """
        # Get the base configuration
        base_config = super().get_config()

        # Add the non-serializable arguments to the configuration
        base_config.update(
            {
                "learning_rate_dict": self.learning_rate_dict,
                "loss_function": self.loss_function,
                "input_tensors_list": self.input_tensors_list,
                "orig_factor_matrices": self.orig_factor_matrices,
                "force_function_list": self.force_function_list,
                "params_dict": self.params_dict,
                "use_attention": self.use_attention,
                "activation": self.activation,
                "hessian": self.hessian,
                "layer_dims": self.layer_dims,
                "tensor_dtype": self.tensor_dtype,
                "sensor_list": self.sensor_list,
            }
        )

        return base_config

    @tf.function
    def train_step(self, beta=10, bilinear_params_dict=None):  # pragma: no cover
        """
        The train step method for the model.

        :param beta: The beta parameter for the training step, defaults to 10.
        :type beta: int, optional
        :param bilinear_params_dict: The dictionary containing the bilinear parameters, defaults to None.
        :type bilinear_params_dict: dict, optional
        :return: The output of the training step.
        :rtype: varies based on implementation
        """

        with tf.GradientTape(persistent=True) as tape:
            # Predict the values for dirichlet boundary conditions
            predicted_values_dirichlet = self(self.dirichlet_input)
            # reshape the predicted values to (, 1)
            predicted_values_dirichlet = tf.reshape(predicted_values_dirichlet[:, 0], [-1, 1])

            # predict the sensor values
            predicted_sensor_values = self(self.sensor_points)
            # reshape the predicted values to (, 1)
            predicted_sensor_values = tf.reshape(predicted_sensor_values[:, 0], [-1, 1])

            # initialize total loss as a tensor with shape (1,) and value 0.0
            total_pde_loss = 0.0

            with tf.GradientTape(persistent=True) as tape1:
                # tape gradient
                tape1.watch(self.input_tensor)
                # Compute the predicted values from the model
                predicted_values_actual = self(self.input_tensor)

                predicted_values = predicted_values_actual[:, 0]
                inverse_param_values = predicted_values_actual[:, 1]

            # compute the gradients of the predicted values wrt the input which is (x, y)
            # First column of the predicted values is the predicted value of the PDE
            gradients = tape1.gradient(predicted_values, self.input_tensor)

            # obtain inverse param gradients
            inverse_param_gradients = tape1.gradient(inverse_param_values, self.input_tensor)

            # Split the gradients into x and y components and reshape them to (-1, 1)
            # the reshaping is done for the tensorial operations purposes (refer Notebook)
            pred_grad_x = tf.reshape(
                gradients[:, 0], [self.n_cells, self.pre_multiplier_grad_x.shape[-1]]
            )  # shape : (N_cells , N_quadrature_points)
            pred_grad_y = tf.reshape(
                gradients[:, 1], [self.n_cells, self.pre_multiplier_grad_y.shape[-1]]
            )  # shape : (N_cells , N_quadrature_points)

            # First column of the predicted values is the predicted value of the PDE and reshape it to (N_cells, N_quadrature_points)
            pred_val = tf.reshape(
                predicted_values, [self.n_cells, self.pre_multiplier_val.shape[-1]]
            )  # shape : (N_cells , N_quadrature_points)

            # reshape the second column of the predicted value and reshape it to (N_cells, N_quadrature_points)
            inverse_param_values = tf.reshape(
                inverse_param_values, [self.n_cells, self.pre_multiplier_val.shape[-1]]
            )  # shape : (N_cells , N_quadrature_points)

            cells_residual = self.loss_function(
                test_shape_val_mat=self.pre_multiplier_val,
                test_grad_x_mat=self.pre_multiplier_grad_x,
                test_grad_y_mat=self.pre_multiplier_grad_y,
                pred_nn=pred_val,
                pred_grad_x_nn=pred_grad_x,
                pred_grad_y_nn=pred_grad_y,
                forcing_function=self.force_matrix,
                bilinear_params=bilinear_params_dict,
                inverse_params_list=[inverse_param_values],
            )

            residual = tf.reduce_sum(cells_residual)

            # Compute the total loss for the PDE
            total_pde_loss = total_pde_loss + residual

            # print shapes of the predicted values and the actual values
            boundary_loss = tf.reduce_mean(
                tf.square(predicted_values_dirichlet - self.dirichlet_actual), axis=0
            )

            # Sensor loss
            sensor_loss = tf.reduce_mean(
                tf.square(predicted_sensor_values - self.sensor_values), axis=0
            )

            # Compute Total Loss
            total_loss = total_pde_loss + beta * boundary_loss + 10 * sensor_loss

        trainable_vars = self.trainable_variables
        self.gradients = tape.gradient(total_loss, trainable_vars)
        self.optimizer.apply_gradients(zip(self.gradients, trainable_vars))

        return {
            "loss_pde": total_pde_loss,
            "loss_dirichlet": boundary_loss,
            "loss": total_loss,
            "sensor_loss": sensor_loss,
        }



"""
The file `plot_utils.py` contains the utility functions for
plotting the loss functions and the predicted inverse parameters

Author: Thivin Anandh D

Date: 02/Nov/2023

Changelog: 02/Nov/2023 - file is created; and added functions to plot the loss functions and the predicted inverse parameters

Known issues: None
"""

import matplotlib.pyplot as plt
from cycler import cycler
import numpy as np


plt.rcParams["xtick.labelsize"] = 20
plt.rcParams["axes.titlesize"] = 20
plt.rcParams["axes.labelsize"] = 20

plt.rcParams["legend.fontsize"] = 20
plt.rcParams["ytick.labelsize"] = 20
plt.rcParams["axes.prop_cycle"] = cycler(
    color=[
        "darkblue",
        "#d62728",
        "#2ca02c",
        "#ff7f0e",
        "#bcbd22",
        "#8c564b",
        "#17becf",
        "#9467bd",
        "#e377c2",
        "#7f7f7f",
    ]
)


# plot the loss function
def plot_loss_function(loss_function, output_path):
    """
    This function will plot the loss function.

    :param loss_function: list of loss values
    :type loss_function: list
    :param output_path: path to save the plot
    :type output_path: str

    :return: None
    :rtype: None
    """
    # plot the loss function
    plt.figure(figsize=(6.4, 4.8))
    plt.plot(loss_function)
    # plot y axis in log scale
    plt.yscale("log")
    plt.xlabel("Epochs")
    plt.ylabel("Loss")
    plt.title("Loss Function")
    plt.tight_layout()
    plt.grid()
    plt.savefig(output_path + "/loss_function.png", dpi=300)

    plt.close()


def plot_array(array, output_path, filename, title, x_label="Epochs", y_label="Loss"):
    """
    This function will plot the loss function.

    :param array: list of loss values
    :type array: list
    :param output_path: path to save the plot
    :type output_path: str
    :param filename: filename to save the plot
    :type filename: str
    :param title: title of the plot
    :type title: str
    :param x_label: x-axis label, defaults to "Epochs"
    :type x_label: str, optional
    :param y_label: y-axis label, defaults to "Loss"
    :type y_label: str, optional

    :return: None
    :rtype: None
    """
    # plot the loss function
    plt.figure(figsize=(6.4, 4.8))
    plt.plot(array)
    # plot y axis in log scale
    plt.yscale("log")
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title)
    plt.tight_layout()
    plt.grid()
    plt.savefig(output_path + f"/{filename}.png", dpi=300)
    plt.close()


# general utility to plot multiple parameters
def plot_multiple_loss_function(
    loss_function_list, output_path, filename, legend_labels, y_label, title, x_label="Epochs"
):
    """
    This function will plot the loss function in log scale for multiple parameters.

    :param loss_function_list: list of loss values for multiple parameters
    :type loss_function_list: list
    :param output_path: path to save the plot
    :type output_path: str
    :param filename: filename to save the plot
    :type filename: str
    :param legend_labels: list of legend labels
    :type legend_labels: list
    :param y_label: y-axis label
    :type y_label: str
    :param title: title of the plot
    :type title: str
    :param x_label: x-axis label, defaults to "Epochs"
    :type x_label: str, optional

    :return: None
    :rtype: None
    """

    # plot the loss function
    plt.figure(figsize=(6.4, 4.8))
    for loss_function, label in zip(loss_function_list, legend_labels):
        plt.plot(loss_function, label=label)

    # plot y axis in log scale
    plt.yscale("log")
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title)
    plt.tight_layout()
    plt.legend()
    plt.grid()
    plt.savefig(output_path + f"/{filename}.png", dpi=300)
    plt.close()


# plot the loss function
def plot_inverse_test_loss_function(loss_function, output_path):
    """
    This function will plot the test loss function of the inverse parameter.

    :param loss_function: list of loss values
    :type loss_function: list
    :param output_path: path to save the plot
    :type output_path: str

    :return: None
    :rtype: None
    """
    # plot the loss function
    plt.figure(figsize=(6.4, 4.8))
    plt.plot(loss_function)
    # plot y axis in log scale
    plt.yscale("log")
    plt.xlabel("Epochs")
    plt.ylabel("Loss")
    plt.title("Loss Function")
    plt.tight_layout()
    plt.savefig(output_path + "/test_inverse_loss_function.png", dpi=300)
    plt.close()


def plot_test_loss_function(loss_function, output_path, fileprefix=""):
    """
    This function will plot the test loss function.

    :param loss_function: list of loss values
    :type loss_function: list
    :param output_path: path to save the plot
    :type output_path: str
    :param fileprefix: prefix for the filename, defaults to ""
    :type fileprefix: str, optional

    :return: None
    :rtype: None
    """
    # plot the loss function
    plt.figure(figsize=(6.4, 4.8))
    plt.plot(loss_function)
    # plot y axis in log scale
    plt.yscale("log")
    plt.xlabel("Epochs")
    plt.ylabel("Loss")
    plt.title("Loss Function")
    plt.tight_layout()
    if fileprefix == "":
        plt.savefig(output_path + "/test_loss_function.png", dpi=300)
    else:
        plt.savefig(output_path + "/" + fileprefix + "_test_loss_function.png", dpi=300)
    plt.close()


def plot_test_time_loss_function(time_array, loss_function, output_path):
    """
    This function will plot the test loss as a function of time in seconds.

    :param time_array: array of time values
    :type time_array: numpy.ndarray
    :param loss_function: list of loss values
    :type loss_function: list
    :param output_path: path to save the plot
    :type output_path: str

    :return: None
    :rtype: None
    """
    # plot the loss function
    plt.figure(figsize=(6.4, 4.8))
    plt.plot(time_array, loss_function)
    # plot y axis in log scale
    plt.yscale("log")
    plt.xscale("log")
    plt.xlabel("Time [s]")
    plt.ylabel("MAE Loss")
    plt.title("Loss Function")
    plt.tight_layout()
    plt.savefig(output_path + "/test_time_loss_function.png", dpi=300)
    plt.close()


def plot_contour(x, y, z, output_path, filename, title):
    """
    This function will plot the contour plot.

    :param x: x values
    :type x: numpy.ndarray
    :param y: y values
    :type y: numpy.ndarray
    :param z: z values
    :type z: numpy.ndarray
    :param output_path: path to save the plot
    :type output_path: str
    :param filename: filename to save the plot
    :type filename: str
    :param title: title of the plot
    :type title: str

    :return: None
    :rtype: None
    """

    plt.figure(figsize=(6.4, 4.8))
    plt.contourf(x, y, z, levels=100, cmap="jet")
    plt.title(title)
    plt.colorbar()
    plt.savefig(output_path + "/" + filename + ".png", dpi=300)

    plt.close()


# plot the Inverse parameter prediction
def plot_inverse_param_function(
    inverse_predicted, inverse_param_name, actual_value, output_path, file_prefix
):
    """
    This function will plot the predicted inverse parameter.

    :param inverse_predicted: list of predicted inverse parameter values
    :type inverse_predicted: list
    :param inverse_param_name: name of the inverse parameter
    :type inverse_param_name: str
    :param actual_value: actual value of the inverse parameter
    :type actual_value: float
    :param output_path: path to save the plot
    :type output_path: str
    :param file_prefix: prefix for the filename
    :type file_prefix: str

    :return: None
    :rtype: None
    """
    # plot the loss function
    plt.figure(figsize=(6.4, 4.8), dpi=300)
    plt.plot(inverse_predicted, label="Predicted " + inverse_param_name)

    # draw a horizontal dotted line at the actual value
    plt.hlines(
        actual_value,
        0,
        len(inverse_predicted),
        colors="k",
        linestyles="dashed",
        label="Actual " + inverse_param_name,
    )

    # plot y axis in log scale
    # plt.yscale("log")
    plt.xlabel("Epochs")
    plt.ylabel(inverse_param_name)

    # plt.title("Loss Function")
    plt.tight_layout()
    plt.legend()
    plt.grid()
    plt.savefig(output_path + f"/{file_prefix}.png", dpi=300)
    plt.close()

    # plot the loss of inverse parameter
    plt.figure(figsize=(6.4, 4.8), dpi=300)
    actual_val_array = np.ones_like(inverse_predicted) * actual_value
    plt.plot(abs(actual_val_array - inverse_predicted))
    plt.xlabel("Epochs")
    plt.ylabel("Absolute Error")
    plt.yscale("log")
    plt.title("Absolute Error of " + inverse_param_name)
    plt.tight_layout()
    plt.savefig(output_path + f"/{file_prefix}_absolute_error.png", dpi=300)
    plt.close()



"""
The file `quad_affine.py` defines the Quad Affine transformation of the reference element.
The implementation is referenced from the ParMooN project  (File: QuadAffine.C).

Author: Thivin Anandh D

Changelog: 30/Aug/2023 - Initial version

Known issues: None

Dependencies: None specified
"""

import numpy as np
from .fe_transformation_2d import FETransforamtion2D


class QuadAffin(FETransforamtion2D):
    """
    Defines the Quad Affine transformation of the reference element.

    :param co_ordinates: The coordinates of the reference element.
    :type co_ordinates: numpy.ndarray
    """

    def __init__(self, co_ordinates) -> None:
        """
        Constructor for the QuadAffin class.

        :param co_ordinates: The coordinates of the reference element.
        :type co_ordinates: numpy.ndarray
        """
        self.co_ordinates = co_ordinates
        self.set_cell()
        self.get_jacobian(
            0, 0
        )  # 0,0 is just a dummy value # this sets the jacobian and the inverse of the jacobian

    def set_cell(self):
        """
        Set the cell coordinates, which will be used to calculate the Jacobian and actual values.

        :param None:
            There are no parameters for this method.

        :returns None:
            This method does not return anything.
        """

        self.x0 = self.co_ordinates[0][0]
        self.x1 = self.co_ordinates[1][0]
        self.x2 = self.co_ordinates[2][0]
        self.x3 = self.co_ordinates[3][0]

        # get the y-coordinates of the cell
        self.y0 = self.co_ordinates[0][1]
        self.y1 = self.co_ordinates[1][1]
        self.y2 = self.co_ordinates[2][1]
        self.y3 = self.co_ordinates[3][1]

        self.xc0 = (self.x1 + self.x3) * 0.5
        self.xc1 = (self.x1 - self.x0) * 0.5
        self.xc2 = (self.x3 - self.x0) * 0.5

        self.yc0 = (self.y1 + self.y3) * 0.5
        self.yc1 = (self.y1 - self.y0) * 0.5
        self.yc2 = (self.y3 - self.y0) * 0.5

    def get_original_from_ref(self, xi, eta):
        """
        Returns the original coordinates from the reference coordinates.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: numpy.ndarray
            The original coordinates.
        """
        x = self.xc0 + self.xc1 * xi + self.xc2 * eta
        y = self.yc0 + self.yc1 * xi + self.yc2 * eta

        return np.array([x, y])

    def get_jacobian(self, xi, eta):
        """
        Returns the Jacobian of the transformation.

        :param xi: The xi coordinate.
        :type xi: float
        :param eta: The eta coordinate.
        :type eta: float

        :return: The Jacobian of the transformation.
        :rtype: float
        """
        self.detjk = self.xc1 * self.yc2 - self.xc2 * self.yc1
        self.rec_detjk = 1 / self.detjk

        return abs(self.detjk)

    def get_orig_from_ref_derivative(self, ref_gradx, ref_grady, xi, eta):
        """
        Returns the derivatives of the original coordinates with respect to the reference coordinates.

        :param ref_gradx: The reference gradient in the x-direction.
        :type ref_gradx: numpy.ndarray
        :param ref_grady: The reference gradient in the y-direction.
        :type ref_grady: numpy.ndarray
        :param xi: The xi coordinate.
        :type xi: float
        :param eta: The eta coordinate.
        :type eta: float

        :return: The derivatives of the original coordinates with respect to the reference coordinates.
        :rtype: tuple
        """
        gradx_orig = np.zeros(ref_gradx.shape)
        grady_orig = np.zeros(ref_grady.shape)

        for i in range(ref_gradx.shape[0]):
            gradx_orig[i] = (self.yc2 * ref_gradx[i] - self.yc1 * ref_grady[i]) * self.rec_detjk
            grady_orig[i] = (-self.xc2 * ref_gradx[i] + self.xc1 * ref_grady[i]) * self.rec_detjk

        return gradx_orig, grady_orig

    def get_orig_from_ref_second_derivative(self, grad_xx_ref, grad_xy_ref, grad_yy_ref, xi, eta):
        """
        Returns the second derivatives (xx, xy, yy) of the original coordinates with respect to the reference coordinates.

        :param grad_xx_ref: The reference second derivative in the xx-direction.
        :type grad_xx_ref: numpy.ndarray
        :param grad_xy_ref: The reference second derivative in the xy-direction.
        :type grad_xy_ref: numpy.ndarray
        :param grad_yy_ref: The reference second derivative in the yy-direction.
        :type grad_yy_ref: numpy.ndarray
        :param xi: The xi coordinate.
        :type xi: float
        :param eta: The eta coordinate.
        :type eta: float

        :return: The second derivatives (xx, xy, yy) of the original coordinates with respect to the reference coordinates.
        :rtype: tuple
        """
        GeoData = np.zeros((3, 3))
        Eye = np.identity(3)

        # Populate GeoData (assuming xc1, xc2, yc1, yc2 are defined)
        GeoData[0, 0] = self.xc1 * self.xc1
        GeoData[0, 1] = 2 * self.xc1 * self.yc1
        GeoData[0, 2] = self.yc1 * self.yc1
        GeoData[1, 0] = self.xc1 * self.xc2
        GeoData[1, 1] = self.yc1 * self.xc2 + self.xc1 * self.yc2
        GeoData[1, 2] = self.yc1 * self.yc2
        GeoData[2, 0] = self.xc2 * self.xc2
        GeoData[2, 1] = 2 * self.xc2 * self.yc2
        GeoData[2, 2] = self.yc2 * self.yc2

        # solve the linear system
        solution = np.linalg.solve(GeoData, Eye)

        # generate empty arrays for the original second derivatives
        grad_xx_orig = np.zeros(grad_xx_ref.shape)
        grad_xy_orig = np.zeros(grad_xy_ref.shape)
        grad_yy_orig = np.zeros(grad_yy_ref.shape)

        for j in range(grad_xx_ref.shape[0]):
            r20 = grad_xx_ref[j]
            r11 = grad_xy_ref[j]
            r02 = grad_yy_ref[j]

            grad_xx_orig[j] = solution[0, 0] * r20 + solution[0, 1] * r11 + solution[0, 2] * r02
            grad_xy_orig[j] = solution[1, 0] * r20 + solution[1, 1] * r11 + solution[1, 2] * r02
            grad_yy_orig[j] = solution[2, 0] * r20 + solution[2, 1] * r11 + solution[2, 2] * r02

        return grad_xx_orig, grad_xy_orig, grad_yy_orig



"""
This file `datahandler2d.py` is to handle data for 2D problems, convert them into tensors using custom tf functions
and make them available for the model to train

Author : Thivin Anandh D

Date : 22/Sep/2023

History : 22/Sep/2023 - Initial implementation with basic data handling
"""

from ..FE.fespace2d import *
from ..Geometry.geometry_2d import *
import tensorflow as tf

from .datahandler import DataHandler


class DataHandler2D(DataHandler):
    """
    This class is to handle data for 2D problems, convert them into tensors using custom tf functions.
    It is responsible for all type conversions and data handling.

    .. note:: All inputs to these functions are generally numpy arrays with dtype np.float64.
              So we can either maintain the same dtype or convert them to tf.float32 ( for faster computation ).

    :param fespace: The FESpace2D object.
    :type fespace: FESpace2D
    :param domain: The Domain2D object.
    :type domain: Domain2D
    :param shape_val_mat_list: List of shape function values for each cell.
    :type shape_val_mat_list: list
    :param grad_x_mat_list: List of shape function derivatives with respect to x for each cell.
    :type grad_x_mat_list: list
    :param grad_y_mat_list: List of shape function derivatives with respect to y for each cell.
    :type grad_y_mat_list: list
    :param x_pde_list: List of actual coordinates of the quadrature points for each cell.
    :type x_pde_list: list
    :param forcing_function_list: List of forcing function values for each cell.
    :type forcing_function_list: list
    :param dtype: The tensorflow dtype to be used for all the tensors.
    :type dtype: tf.DType
    """

    def __init__(self, fespace, domain, dtype):
        """
        Constructor for the DataHandler2D class

        :param fespace: The FESpace2D object.
        :type fespace: FESpace2D
        :param domain: The Domain2D object.
        :type domain: Domain2D
        :param shape_val_mat_list: List of shape function values for each cell.
        :type shape_val_mat_list: list
        :param grad_x_mat_list: List of shape function derivatives with respect to x for each cell.
        :type grad_x_mat_list: list
        :param grad_y_mat_list: List of shape function derivatives with respect to y for each cell.
        :type grad_y_mat_list: list
        :param x_pde_list: List of actual coordinates of the quadrature points for each cell.
        :type x_pde_list: list
        :param forcing_function_list: List of forcing function values for each cell.
        :type forcing_function_list: list
        :param dtype: The tensorflow dtype to be used for all the tensors.
        :type dtype: tf.DType
        """
        # call the parent class constructor
        super().__init__(fespace=fespace, domain=domain, dtype=dtype)

        self.shape_val_mat_list = []
        self.grad_x_mat_list = []
        self.grad_y_mat_list = []
        self.x_pde_list = []
        self.forcing_function_list = []

        # check if the given dtype is a valid tensorflow dtype
        if not isinstance(self.dtype, tf.DType):
            raise TypeError("The given dtype is not a valid tensorflow dtype")

        for cell_index in range(self.fespace.n_cells):
            shape_val_mat = tf.constant(
                self.fespace.get_shape_function_val(cell_index), dtype=self.dtype
            )
            grad_x_mat = tf.constant(
                self.fespace.get_shape_function_grad_x(cell_index), dtype=self.dtype
            )
            grad_y_mat = tf.constant(
                self.fespace.get_shape_function_grad_y(cell_index), dtype=self.dtype
            )
            x_pde = tf.constant(
                self.fespace.get_quadrature_actual_coordinates(cell_index), dtype=self.dtype
            )
            forcing_function = tf.constant(
                self.fespace.get_forcing_function_values(cell_index), dtype=self.dtype
            )
            self.shape_val_mat_list.append(shape_val_mat)
            self.grad_x_mat_list.append(grad_x_mat)
            self.grad_y_mat_list.append(grad_y_mat)
            self.x_pde_list.append(x_pde)
            self.forcing_function_list.append(forcing_function)

        # now convert all the shapes into 3D tensors for easy multiplication
        # input tensor - x_pde_list
        self.x_pde_list = tf.reshape(self.x_pde_list, [-1, 2])

        self.forcing_function_list = tf.concat(self.forcing_function_list, axis=1)

        self.shape_val_mat_list = tf.stack(self.shape_val_mat_list, axis=0)
        self.grad_x_mat_list = tf.stack(self.grad_x_mat_list, axis=0)
        self.grad_y_mat_list = tf.stack(self.grad_y_mat_list, axis=0)

        # test points
        self.test_points = None

    def get_dirichlet_input(self):
        """
        This function will return the input for the Dirichlet boundary data

        :return:
            - input_dirichlet (tf.Tensor): The input for the Dirichlet boundary data
            - actual_dirichlet (tf.Tensor): The actual Dirichlet boundary data
        """
        input_dirichlet, actual_dirichlet = self.fespace.generate_dirichlet_boundary_data()

        # convert to tensors
        input_dirichlet = tf.constant(input_dirichlet, dtype=self.dtype)
        actual_dirichlet = tf.constant(actual_dirichlet, dtype=self.dtype)
        actual_dirichlet = tf.reshape(actual_dirichlet, [-1, 1])

        return input_dirichlet, actual_dirichlet

    def get_test_points(self):
        """
        Get the test points for the given domain.

        :return: The test points for the given domain.
        :rtype: tf.Tensor
        """
        self.test_points = self.domain.get_test_points()
        self.test_points = tf.constant(self.test_points, dtype=self.dtype)
        return self.test_points

    def get_bilinear_params_dict_as_tensors(self, function):
        """
        Accepts a function from example file and converts all the values into tensors of the given dtype

        Parameters:
        - function (function): The function from the example file which returns the bilinear parameters dictionary

        Returns:
        - bilinear_params_dict (dict): The bilinear parameters dictionary with all the values converted to tensors

        :param function: The function from the example file which returns the bilinear parameters dictionary
        :type function: function
        :return: The bilinear parameters dictionary with all the values converted to tensors
        :rtype: dict
        """
        # get the dictionary of bilinear parameters
        bilinear_params_dict = function()

        # loop over all keys and convert the values to tensors
        for key in bilinear_params_dict.keys():
            bilinear_params_dict[key] = tf.constant(bilinear_params_dict[key], dtype=self.dtype)

        return bilinear_params_dict

    # to be used only in inverse problems
    def get_sensor_data(self, exact_sol, num_sensor_points, mesh_type, file_name=None):
        """
        Accepts a function from example file and converts all the values into tensors of the given dtype

        :param exact_sol: The function from the example file which returns the exact solution
        :type exact_sol: function
        :param num_sensor_points: The number of sensor points to be generated
        :type num_sensor_points: int
        :param mesh_type: The type of mesh to be used for sensor data generation
        :type mesh_type: str
        :param file_name: The name of the file to be used for external mesh generation, defaults to None
        :type file_name: str, optional
        :return: The sensor points and sensor values as tensors
        :rtype: tuple[tf.Tensor, tf.Tensor]
        """
        print(f"mesh_type = {mesh_type}")
        if mesh_type == "internal":
            # Call the method to get the sensor data
            points, sensor_values = self.fespace.get_sensor_data(exact_sol, num_sensor_points)
        elif mesh_type == "external":
            # Call the method to get the sensor data
            points, sensor_values = self.fespace.get_sensor_data_external(
                exact_sol, num_sensor_points, file_name
            )
        # convert the points and sensor values into tensors
        points = tf.constant(points, dtype=self.dtype)
        sensor_values = tf.constant(sensor_values, dtype=self.dtype)

        sensor_values = tf.reshape(sensor_values, [-1, 1])
        points = tf.reshape(points, [-1, 2])

        return points, sensor_values

    # get inverse param dict as tensors
    def get_inverse_params(self, inverse_params_dict_function):
        """
        Accepts a function from example file and converts all the values into tensors of the given dtype

        :param inverse_params_dict_function: The function from the example file which returns the inverse parameters dictionary
        :type inverse_params_dict_function: function
        :return: The inverse parameters dictionary with all the values converted to tensors
        :rtype: dict
        """
        # loop over all keys and convert the values to tensors

        inverse_params_dict = inverse_params_dict_function()

        for key in inverse_params_dict.keys():
            inverse_params_dict[key] = tf.constant(inverse_params_dict[key], dtype=self.dtype)

        return inverse_params_dict



"""
This file `poisson2d_inverse.py` is implemntation of our efficient tensor-based loss calculation for poisson equation with inverse problem (constant).

Author: Thivin Anandh D

Date: 21/Sep/2023

History: Initial implementation

Refer: https://arxiv.org/abs/2404.12063
"""

import tensorflow as tf


# PDE loss function for the poisson problem inverse
# @tf.function - Commented due to compatibility issues
def pde_loss_poisson_inverse(
    test_shape_val_mat,
    test_grad_x_mat,
    test_grad_y_mat,
    pred_nn,
    pred_grad_x_nn,
    pred_grad_y_nn,
    forcing_function,
    bilinear_params,
    inverse_params_dict,
):  # pragma: no cover
    """
    Calculates and returns the loss for the  Poisson problem Inverse (constant)

    :param test_shape_val_mat: The test shape value matrix.
    :type test_shape_val_mat: tf.Tensor
    :param test_grad_x_mat: The x-gradient of the test matrix.
    :type test_grad_x_mat: tf.Tensor
    :param test_grad_y_mat: The y-gradient of the test matrix.
    :type test_grad_y_mat: tf.Tensor
    :param pred_nn: The predicted neural network output.
    :type pred_nn: tf.Tensor
    :param pred_grad_x_nn: The x-gradient of the predicted neural network output.
    :type pred_grad_x_nn: tf.Tensor
    :param pred_grad_y_nn: The y-gradient of the predicted neural network output.
    :type pred_grad_y_nn: tf.Tensor
    :param forcing_function: The forcing function used in the PDE.
    :type forcing_function: function
    :param bilinear_params_dict: The dictionary containing the bilinear parameters.
    :type bilinear_params_dict: dict
    :param inverse_param_dict: The dictionary containing the parameters for the inverse problem.
    :type inverse_param_dict: dict

    :return: The calculated loss.
    :rtype: tf.Tensor
    """
    # ∫du/dx. dv/dx dΩ
    pde_diffusion_x = tf.transpose(tf.linalg.matvec(test_grad_x_mat, pred_grad_x_nn))

    # ∫du/dy. dv/dy dΩ
    pde_diffusion_y = tf.transpose(tf.linalg.matvec(test_grad_y_mat, pred_grad_y_nn))

    # eps * ∫ (du/dx. dv/dx + du/dy. dv/dy) dΩ
    pde_diffusion = inverse_params_dict["eps"] * (pde_diffusion_x + pde_diffusion_y)

    residual_matrix = pde_diffusion - forcing_function

    residual_cells = tf.reduce_mean(tf.square(residual_matrix), axis=0)

    return residual_cells



"""
file: basis_2d_QN_Chebyshev_2.py
description: This file contains the class Basis2DQNChebyshev2 which defines the basis functions for a 
              2D Q1 element using Chebyshev polynomials.
              Test functions and derivatives are inferred from the work by Ehsan Kharazmi et.al
             (hp-VPINNs: Variational Physics-Informed Neural Networks With Domain Decomposition)
             available at https://github.com/ehsankharazmi/hp-VPINNs/
authors: Thivin Anandh D
changelog: 30/Aug/2023 - Initial version
known_issues: None
"""

# import the legendre polynomials
from scipy.special import jacobi

import numpy as np
from .basis_function_2d import BasisFunction2D


class Basis2DQNChebyshev2(BasisFunction2D):
    """
    This class defines the basis functions for a 2D Q1 element.
    """

    def __init__(self, num_shape_functions: int):
        super().__init__(num_shape_functions)

    def jacobi_wrapper(self, n, a, b, x):
        """
        Evaluate the Jacobi polynomial of degree n with parameters a and b at the given points x.

        :param n: Degree of the Jacobi polynomial.
        :type n: int
        :param a: First parameter of the Jacobi polynomial.
        :type a: float
        :param b: Second parameter of the Jacobi polynomial.
        :type b: float
        :param x: Points at which to evaluate the Jacobi polynomial.
        :type x: array_like

        :return: Values of the Jacobi polynomial at the given points x.
        :rtype: array_like
        """
        x = np.array(x, dtype=np.float64)
        return jacobi(n, a, b)(x)

    ## Helper Function
    def test_fcnx(self, n_test, x):
        """
        Compute the x-component of the test functions for a given number of test functions and x-coordinates.

        :param n_test: Number of test functions.
        :type n_test: int
        :param x: x-coordinates at which to evaluate the test functions.
        :type x: array_like

        :return: Values of the x-component of the test functions.
        :rtype: array_like
        """
        test_total = []
        for n in range(1, n_test + 1):
            test = self.jacobi_wrapper(n + 1, -1 / 2, -1 / 2, x) / self.jacobi_wrapper(
                n + 1, -1 / 2, -1 / 2, 1
            ) - self.jacobi_wrapper(n - 1, -1 / 2, -1 / 2, x) / self.jacobi_wrapper(
                n - 1, -1 / 2, -1 / 2, 1
            )
            test_total.append(test)
        return np.asarray(test_total, np.float64)

    def test_fcny(self, n_test, y):
        """
        Compute the y-component of the test functions for a given number of test functions and y-coordinates.

        :param n_test: Number of test functions.
        :type n_test: int
        :param y: y-coordinates at which to evaluate the test functions.
        :type y: array_like
        :return: Values of the y-component of the test functions.
        :rtype: array_like
        """
        test_total = []
        for n in range(1, n_test + 1):
            test = self.jacobi_wrapper(n + 1, -1 / 2, -1 / 2, y) / self.jacobi_wrapper(
                n + 1, -1 / 2, -1 / 2, 1
            ) - self.jacobi_wrapper(n - 1, -1 / 2, -1 / 2, y) / self.jacobi_wrapper(
                n - 1, -1 / 2, -1 / 2, 1
            )
            test_total.append(test)
        return np.asarray(test_total, np.float64)

    def dtest_fcn(self, n_test, x):
        """
        Compute the first and second derivatives of the test function.

        :param n_test: The number of test functions.
        :type n_test: int
        :param x: The input value.
        :type x: float
        :return: Array of first derivatives of the test function, Array of second derivatives of the test function.
        :rtype: tuple(ndarray, ndarray)
        """
        d1test_total = []
        d2test_total = []
        for n in range(1, n_test + 1):
            if n == 1:
                d1test = (
                    ((n + 1) / 2)
                    * self.jacobi_wrapper(n, 1 / 2, 1 / 2, x)
                    / self.jacobi_wrapper(n + 1, -1 / 2, -1 / 2, 1)
                )
                d2test = (
                    ((n + 2) * (n + 1) / (2 * 2))
                    * self.jacobi_wrapper(n - 1, 3 / 2, 3 / 2, x)
                    / self.jacobi_wrapper(n + 1, -1 / 2, -1 / 2, 1)
                )
                d1test_total.append(d1test)
                d2test_total.append(d2test)
            elif n == 2:
                d1test = ((n + 1) / 2) * self.jacobi_wrapper(
                    n, 1 / 2, 1 / 2, x
                ) / self.jacobi_wrapper(n + 1, -1 / 2, -1 / 2, 1) - (
                    (n - 1) / 2
                ) * self.jacobi_wrapper(
                    n - 2, 1 / 2, 1 / 2, x
                ) / self.jacobi_wrapper(
                    n - 1, -1 / 2, -1 / 2, 1
                )
                d2test = (
                    ((n + 2) * (n + 1) / (2 * 2))
                    * self.jacobi_wrapper(n - 1, 3 / 2, 3 / 2, x)
                    / self.jacobi_wrapper(n + 1, -1 / 2, -1 / 2, 1)
                )
                d1test_total.append(d1test)
                d2test_total.append(d2test)
            else:
                d1test = ((n + 1) / 2) * self.jacobi_wrapper(
                    n, 1 / 2, 1 / 2, x
                ) / self.jacobi_wrapper(n + 1, -1 / 2, -1 / 2, 1) - (
                    (n - 1) / 2
                ) * self.jacobi_wrapper(
                    n - 2, 1 / 2, 1 / 2, x
                ) / self.jacobi_wrapper(
                    n - 1, -1 / 2, -1 / 2, 1
                )
                d2test = ((n + 2) * (n + 1) / (2 * 2)) * self.jacobi_wrapper(
                    n - 1, 3 / 2, 3 / 2, x
                ) / self.jacobi_wrapper(n + 1, -1 / 2, -1 / 2, 1) - (
                    (n) * (n - 1) / (2 * 2)
                ) * self.jacobi_wrapper(
                    n - 3, 3 / 2, 3 / 2, x
                ) / self.jacobi_wrapper(
                    n - 1, -1 / 2, -1 / 2, 1
                )
                d1test_total.append(d1test)
                d2test_total.append(d2test)
        return np.asarray(d1test_total), np.asarray(d2test_total)

    def value(self, xi, eta):
        """
        This method returns the values of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like
        :param eta: y-coordinates at which to evaluate the basis functions.
        :type eta: array_like
        :return: Values of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        test_x = self.test_fcnx(num_shape_func_in_1d, xi)
        test_y = self.test_fcny(num_shape_func_in_1d, eta)
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                test_x[i, :] * test_y
            )

        return values

    def gradx(self, xi, eta):
        """
        This method returns the x-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like
        :param eta: y-coordinates at which to evaluate the basis functions.
        :type eta: array_like
        :return: Values of the x-derivatives of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        grad_test_x = self.dtest_fcn(num_shape_func_in_1d, xi)[0]
        test_y = self.test_fcny(num_shape_func_in_1d, eta)
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                grad_test_x[i, :] * test_y
            )

        return values

    def grady(self, xi, eta):
        """
        This method returns the y-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like
        :param eta: y-coordinates at which to evaluate the basis functions.

        :return: Values of the y-derivatives of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        test_x = self.test_fcnx(num_shape_func_in_1d, xi)
        grad_test_y = self.dtest_fcn(num_shape_func_in_1d, eta)[0]
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                test_x[i, :] * grad_test_y
            )

        return values

    def gradxx(self, xi, eta):
        """
        This method returns the xx-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like
        :param eta: y-coordinates at which to evaluate the basis functions.
        :type eta: array_like

        :return: Values of the xx-derivatives of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        grad_grad_x = self.dtest_fcn(num_shape_func_in_1d, xi)[1]
        test_y = self.test_fcny(num_shape_func_in_1d, eta)
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                grad_grad_x[i, :] * test_y
            )

        return values

    def gradxy(self, xi, eta):
        """
        This method returns the xy-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like
        :param eta: y-coordinates at which to evaluate the basis functions.
        :type eta: array_like
        :return: Values of the xy-derivatives of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        grad_test_x = self.dtest_fcn(num_shape_func_in_1d, xi)[0]
        grad_test_y = self.dtest_fcn(num_shape_func_in_1d, eta)[0]
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                grad_test_x[i, :] * grad_test_y
            )

        return values

    def gradyy(self, xi, eta):
        """
        This method returns the yy-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: x-coordinates at which to evaluate the basis functions.
        :type xi: array_like
        :param eta: y-coordinates at which to evaluate the basis functions.
        :type eta: array_like

        :return: Values of the yy-derivatives of the basis functions.
        :rtype: array_like
        """
        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))
        test_x = self.test_fcnx(num_shape_func_in_1d, xi)
        grad_grad_y = self.dtest_fcn(num_shape_func_in_1d, eta)[1]
        values = np.zeros((self.num_shape_functions, len(xi)), dtype=np.float64)

        for i in range(num_shape_func_in_1d):
            values[num_shape_func_in_1d * i : num_shape_func_in_1d * (i + 1), :] = (
                test_x[i, :] * grad_grad_y
            )

        return values



"""
The file `fespace2d.py` contains the main class that holds the information of all the 
Finite Element (FE) spaces of all the cells within the given mesh.

Author: Thivin Anandh D

Changelog: 30/Aug/2023 - Initial version

Known issues: None

Dependencies: None specified
"""

import numpy as np
import meshio
from .FE2D_Cell import FE2D_Cell

# from rich.progress import Progress, TextColumn, BarColumn, TimeElapsedColumn
from tqdm import tqdm

# import plotting
import matplotlib.pyplot as plt

# import path
from pathlib import Path

# import tensorflow
import tensorflow as tf

from ..utils.print_utils import print_table

from pyDOE import lhs
import pandas as pd

from matplotlib import rc
from cycler import cycler


plt.rcParams["xtick.labelsize"] = 20
plt.rcParams["axes.titlesize"] = 20
plt.rcParams["axes.labelsize"] = 20

plt.rcParams["legend.fontsize"] = 20
plt.rcParams["ytick.labelsize"] = 20
plt.rcParams["axes.prop_cycle"] = cycler(
    color=[
        "darkblue",
        "#d62728",
        "#2ca02c",
        "#ff7f0e",
        "#bcbd22",
        "#8c564b",
        "#17becf",
        "#9467bd",
        "#e377c2",
        "#7f7f7f",
    ]
)

from .fespace import Fespace


class Fespace2D(Fespace):
    """
    Represents a finite element space in 2D.

    :param mesh: The mesh object.
    :type mesh: Mesh
    :param cells: The array of cell indices.
    :type cells: ndarray
    :param boundary_points: The dictionary of boundary points.
    :type boundary_points: dict
    :param cell_type: The type of cell.
    :type cell_type: str
    :param fe_order: The order of the finite element.
    :type fe_order: int
    :param fe_type: The type of finite element.
    :type fe_type: str
    :param quad_order: The order of the quadrature.
    :type quad_order: int
    :param quad_type: The type of quadrature.
    :type quad_type: str
    :param fe_transformation_type: The type of finite element transformation.
    :type fe_transformation_type: str
    :param bound_function_dict: The dictionary of boundary functions.
    :type bound_function_dict: dict
    :param bound_condition_dict: The dictionary of boundary conditions.
    :type bound_condition_dict: dict
    :param forcing_function: The forcing function.
    :type forcing_function: function
    :param output_path: The output path.
    :type output_path: str
    :param generate_mesh_plot: Whether to generate a plot of the mesh. Defaults to False.
    :type generate_mesh_plot: bool, optional
    """

    def __init__(
        self,
        mesh,
        cells,
        boundary_points,
        cell_type: str,
        fe_order: int,
        fe_type: str,
        quad_order: int,
        quad_type: str,
        fe_transformation_type: str,
        bound_function_dict: dict,
        bound_condition_dict: dict,
        forcing_function,
        output_path: str,
        generate_mesh_plot: bool = False,
    ) -> None:
        """
        The constructor of the Fespace2D class.
        """
        # call the constructor of the parent class
        super().__init__(
            mesh=mesh,
            cells=cells,
            boundary_points=boundary_points,
            cell_type=cell_type,
            fe_order=fe_order,
            fe_type=fe_type,
            quad_order=quad_order,
            quad_type=quad_type,
            fe_transformation_type=fe_transformation_type,
            bound_function_dict=bound_function_dict,
            bound_condition_dict=bound_condition_dict,
            forcing_function=forcing_function,
            output_path=output_path,
        )

        if self.cell_type == "triangle":
            raise ValueError(
                "Triangle Mesh is not supported yet"
            )  # added by thivin - to remove support for triangular mesh

        self.generate_mesh_plot = generate_mesh_plot

        # to be calculated in the plot function
        self.total_dofs = 0
        self.total_boundary_dofs = 0

        # to be calculated on get_boundary_data_dirichlet function
        self.total_dirichlet_dofs = 0

        # get the number of cells
        self.n_cells = self.cells.shape[0]

        self.fe_cell = []

        # Function which assigns the fe_cell for each cell
        self.set_finite_elements()

        # generate the plot of the mesh
        if self.generate_mesh_plot:
            self.generate_plot(self.output_path)
        # self.generate_plot(self.output_path)

        # Obtain boundary Data
        self.dirichlet_boundary_data = self.generate_dirichlet_boundary_data()

        title = [
            "Number of Cells",
            "Number of Quadrature Points",
            "Number of Dirichlet Boundary Points",
            "Quadrature Order",
            "FE Order",
            "FE Type",
            "FE Transformation Type",
        ]
        values = [
            self.n_cells,
            self.total_dofs,
            self.total_dirichlet_dofs,
            self.quad_order,
            self.fe_order,
            self.fe_type,
            self.fe_transformation_type,
        ]
        # print the table
        print_table("FE Space Information", ["Property", "Value"], title, values)

    def set_finite_elements(self) -> None:
        """
        Assigns the finite elements to each cell.

        This method initializes the finite element objects for each cell in the mesh.
        It creates an instance of the `FE2D_Cell` class for each cell, passing the necessary parameters.
        The finite element objects store information about the basis functions, gradients, Jacobians,
        quadrature points, weights, actual coordinates, and forcing functions associated with each cell.

        After initializing the finite element objects, this method prints the shape details of various matrices
        and updates the total number of degrees of freedom (dofs) for the entire mesh.

        :return: None
        """
        progress_bar = tqdm(
            total=self.n_cells,
            desc="Fe2D_cell Setup",
            unit="cells_assembled",
            bar_format="{l_bar}{bar:40}{r_bar}{bar:-10b}",
            colour="blue",
            ncols=100,
        )

        dof = 0
        for i in range(self.n_cells):
            self.fe_cell.append(
                FE2D_Cell(
                    self.cells[i],
                    self.cell_type,
                    self.fe_order,
                    self.fe_type,
                    self.quad_order,
                    self.quad_type,
                    self.fe_transformation_type,
                    self.forcing_function,
                )
            )

            # obtain the shape of the basis function (n_test, N_quad)
            dof += self.fe_cell[i].basis_at_quad.shape[1]

            progress_bar.update(1)
        # print the Shape details of all the matrices from cell 0 using print_table function
        title = [
            "Shape function Matrix Shape",
            "Shape function Gradient Matrix Shape",
            "Jacobian Matrix Shape",
            "Quadrature Points Shape",
            "Quadrature Weights Shape",
            "Quadrature Actual Coordinates Shape",
            "Forcing Function Shape",
        ]
        values = [
            self.fe_cell[0].basis_at_quad.shape,
            self.fe_cell[0].basis_gradx_at_quad.shape,
            self.fe_cell[0].jacobian.shape,
            self.fe_cell[0].quad_xi.shape,
            self.fe_cell[0].quad_weight.shape,
            self.fe_cell[0].quad_actual_coordinates.shape,
            self.fe_cell[0].forcing_at_quad.shape,
        ]
        print_table("FE Matrix Shapes", ["Matrix", "Shape"], title, values)

        # update the total number of dofs
        self.total_dofs = dof

    def generate_plot(self, output_path) -> None:
        """
        Generate a plot of the mesh.

        :param output_path: The path to save the generated plot.
        :type output_path: str
        """
        total_quad = 0
        marker_list = [
            "o",
            ".",
            ",",
            "x",
            "+",
            "P",
            "s",
            "D",
            "d",
            "^",
            "v",
            "<",
            ">",
            "p",
            "h",
            "H",
        ]

        print(f"[INFO] : Generating the plot of the mesh")
        # Plot the mesh
        plt.figure(figsize=(6.4, 4.8), dpi=300)

        # label flag ( to add the label only once)
        label_set = False

        # plot every cell as a quadrilateral
        # loop over all the cells
        for i in range(self.n_cells):
            # get the coordinates of the cell
            x = self.fe_cell[i].cell_coordinates[:, 0]
            y = self.fe_cell[i].cell_coordinates[:, 1]

            # add the first point to the end of the array
            x = np.append(x, x[0])
            y = np.append(y, y[0])

            plt.plot(x, y, "k-", linewidth=0.5)

            # plot the quadrature points
            x_quad = self.fe_cell[i].quad_actual_coordinates[:, 0]
            y_quad = self.fe_cell[i].quad_actual_coordinates[:, 1]

            total_quad += x_quad.shape[0]

            if not label_set:
                plt.scatter(x_quad, y_quad, marker="x", color="b", s=2, label="Quad Pts")
                label_set = True
            else:
                plt.scatter(x_quad, y_quad, marker="x", color="b", s=2)

        self.total_dofs = total_quad

        bound_dof = 0
        # plot the boundary points
        # loop over all the boundary tags
        for i, (bound_id, bound_pts) in enumerate(self.boundary_points.items()):
            # get the coordinates of the boundary points
            x = bound_pts[:, 0]
            y = bound_pts[:, 1]

            # add the first point to the end of the array
            x = np.append(x, x[0])
            y = np.append(y, y[0])

            bound_dof += x.shape[0]

            plt.scatter(x, y, marker=marker_list[i + 1], s=2, label=f"Bd-id : {bound_id}")

        self.total_boundary_dofs = bound_dof

        plt.legend(bbox_to_anchor=(0.85, 1.02))
        plt.axis("equal")
        plt.axis("off")
        plt.tight_layout()

        plt.savefig(str(Path(output_path) / "mesh.png"), bbox_inches="tight")
        plt.savefig(str(Path(output_path) / "mesh.svg"), bbox_inches="tight")

        # print the total number of quadrature points
        print(f"Plots generated")
        print(f"[INFO] : Total number of cells = {self.n_cells}")
        print(f"[INFO] : Total number of quadrature points = {self.total_dofs}")
        print(f"[INFO] : Total number of boundary points = {self.total_boundary_dofs}")

    def generate_dirichlet_boundary_data(self) -> np.ndarray:
        """
        Generate Dirichlet boundary data.

        This function returns the boundary points and their corresponding values.

        :return: A tuple containing two arrays:
            - The first array contains the boundary points as numpy arrays.
            - The second array contains the values of the boundary points as numpy arrays.
        :rtype: Tuple[np.ndarray, np.ndarray]
        """
        x = []
        y = []
        for bound_id, bound_pts in self.boundary_points.items():
            # get the coordinates of the boundary points
            for pt in bound_pts:
                pt_new = np.array([pt[0], pt[1]], dtype=np.float64)
                x.append(pt_new)
                val = np.array(
                    self.bound_function_dict[bound_id](pt[0], pt[1]), dtype=np.float64
                ).reshape(-1, 1)
                y.append(val)

        print(f"[INFO] : Total number of Dirichlet boundary points = {len(x)}")
        self.total_dirichlet_dofs = len(x)
        print(f"[INFO] : Shape of Dirichlet-X = {np.array(x).shape}")
        print(f"[INFO] : Shape of Y = {np.array(y).shape}")

        return x, y

    def generate_dirichlet_boundary_data_vector(self, component) -> np.ndarray:
        """
        Generate the boundary data vector for the Dirichlet boundary condition.

        This function returns the boundary points and their corresponding values for a specific component.

        :param component: The component for which the boundary data vector is generated.
        :type component: int

        :return: The boundary points and their values as numpy arrays.
        :rtype: tuple(numpy.ndarray, numpy.ndarray)
        """
        x = []
        y = []
        for bound_id, bound_pts in self.boundary_points.items():
            # get the coordinates of the boundary points
            for pt in bound_pts:
                pt_new = np.array([pt[0], pt[1]], dtype=np.float64)
                x.append(pt_new)
                val = np.array(
                    self.bound_function_dict[bound_id](pt[0], pt[1])[component], dtype=np.float64
                ).reshape(-1, 1)
                y.append(val)

        return x, y

    def get_shape_function_val(self, cell_index) -> np.ndarray:
        """
        Get the actual values of the shape functions on a given cell.

        :param cell_index: The index of the cell.
        :type cell_index: int

        :return: An array containing the actual values of the shape functions.
        :rtype: np.ndarray

        :raises ValueError: If the cell_index is greater than the number of cells.
        """
        if cell_index >= len(self.fe_cell) or cell_index < 0:
            raise ValueError(
                f"cell_index should be less than {self.n_cells} and greater than or equal to 0"
            )

        return self.fe_cell[cell_index].basis_at_quad.copy()

    def get_shape_function_grad_x(self, cell_index) -> np.ndarray:
        """
        Get the gradient of the shape function with respect to the x-coordinate.

        :param cell_index: The index of the cell.
        :type cell_index: int

        :return: An array containing the gradient of the shape function with respect to the x-coordinate.
        :rtype: np.ndarray

        :raises ValueError: If the cell_index is greater than the number of cells.

        This function returns the actual values of the gradient of the shape function on a given cell.
        """
        if cell_index >= len(self.fe_cell) or cell_index < 0:
            raise ValueError(
                f"cell_index should be less than {self.n_cells} and greater than or equal to 0"
            )

        return self.fe_cell[cell_index].basis_gradx_at_quad.copy()

    def get_shape_function_grad_x_ref(self, cell_index) -> np.ndarray:
        """
        Get the gradient of the shape function with respect to the x-coordinate on the reference element.

        :param cell_index: The index of the cell.
        :type cell_index: int

        :return: An array containing the gradient of the shape function with respect to the x-coordinate.
        :rtype: np.ndarray

        :raises ValueError: If the cell_index is greater than the number of cells.
        """
        if cell_index >= len(self.fe_cell) or cell_index < 0:
            raise ValueError(
                f"cell_index should be less than {self.n_cells} and greater than or equal to 0"
            )

        return self.fe_cell[cell_index].basis_gradx_at_quad_ref.copy()

    def get_shape_function_grad_y(self, cell_index) -> np.ndarray:
        """
        Get the gradient of the shape function with respect to y at the given cell index.

        :param cell_index: The index of the cell.
        :type cell_index: int

        :return: The gradient of the shape function with respect to y.
        :rtype: np.ndarray

        :raises ValueError: If the cell_index is greater than the total number of cells.
        """
        if cell_index >= len(self.fe_cell) or cell_index < 0:
            raise ValueError(
                f"cell_index should be less than {self.n_cells} and greater than or equal to 0"
            )

        return self.fe_cell[cell_index].basis_grady_at_quad.copy()

    def get_shape_function_grad_y_ref(self, cell_index):
        """
        Get the gradient of the shape function with respect to y at the reference element.

        :param cell_index: The index of the cell.
        :type cell_index: int
        :return: The gradient of the shape function with respect to y at the reference element.
        :rtype: np.ndarray
        :raises ValueError: If cell_index is greater than the number of cells.

        This function returns the gradient of the shape function with respect to y at the reference element
        for a given cell. The shape function gradient values are stored in the `basis_grady_at_quad_ref` array
        of the corresponding finite element cell. The `cell_index` parameter specifies the index of the cell
        for which the shape function gradient is required. If the `cell_index` is greater than the total number
        of cells, a `ValueError` is raised.

        .. note::
            The returned gradient values are copied from the `basis_grady_at_quad_ref` array to ensure immutability.
        """
        if cell_index >= len(self.fe_cell) or cell_index < 0:
            raise ValueError(
                f"cell_index should be less than {self.n_cells} and greater than or equal to 0"
            )

        return self.fe_cell[cell_index].basis_grady_at_quad_ref.copy()

    def get_quadrature_actual_coordinates(self, cell_index) -> np.ndarray:
        """
        Get the actual coordinates of the quadrature points for a given cell.

        :param cell_index: The index of the cell.
        :type cell_index: int

        :return: An array containing the actual coordinates of the quadrature points.
        :rtype: np.ndarray

        :raises ValueError: If the cell_index is greater than the number of cells.

        :example:
        >>> fespace = FESpace2D()
        >>> fespace.get_quadrature_actual_coordinates(0)
        array([[0.1, 0.2],
                [0.3, 0.4],
                [0.5, 0.6]])
        """
        if cell_index >= len(self.fe_cell) or cell_index < 0:
            raise ValueError(
                f"cell_index should be less than {self.n_cells} and greater than or equal to 0"
            )

        return self.fe_cell[cell_index].quad_actual_coordinates.copy()

    def get_quadrature_weights(self, cell_index) -> np.ndarray:
        """
        Return the quadrature weights for a given cell.

        :param cell_index: The index of the cell for which the quadrature weights are needed.
        :type cell_index: int
        :return: The quadrature weights for the given cell  of dimension (N_Quad_Points, 1).
        :rtype: np.ndarray
        :raises ValueError: If cell_index is greater than the number of cells.
        Example
        -------
        >>> fespace = FESpace2D()
        >>> weights = fespace.get_quadrature_weights(0)
        >>> print(weights)
        [0.1, 0.2, 0.3, 0.4]
        """
        if cell_index >= len(self.fe_cell) or cell_index < 0:
            raise ValueError(
                f"cell_index should be less than {self.n_cells} and greater than or equal to 0"
            )

        return self.fe_cell[cell_index].mult.copy()

    def get_forcing_function_values(self, cell_index) -> np.ndarray:
        """
        Get the forcing function values at the quadrature points.

        :param cell_index: The index of the cell.
        :type cell_index: int

        :return: The forcing function values at the quadrature points.
        :rtype: np.ndarray

        :raises ValueError: If cell_index is greater than the number of cells.

        This function computes the forcing function values at the quadrature points for a given cell.
        It loops over all the basis functions and computes the integral using the actual coordinates
        and the basis functions at the quadrature points. The resulting values are stored in the
        `forcing_at_quad` attribute of the corresponding `fe_cell` object.

        Note: The forcing function is evaluated using the `forcing_function` method of the `fe_cell`
        object.

        Example usage:
            >>> fespace = FESpace2D()
            >>> cell_index = 0
            >>> forcing_values = fespace.get_forcing_function_values(cell_index)
        """
        if cell_index >= len(self.fe_cell) or cell_index < 0:
            raise ValueError(
                f"cell_index should be less than {self.n_cells} and greater than or equal to 0"
            )

        # Changed by Thivin: To assemble the forcing function at the quadrature points here in the fespace
        # so that it can be used to handle multiple dimensions on a vector valud problem

        # get number of shape functions
        n_shape_functions = self.fe_cell[cell_index].basis_function.num_shape_functions

        # Loop over all the basis functions and compute the integral
        f_integral = np.zeros((n_shape_functions, 1), dtype=np.float64)

        for i in range(n_shape_functions):
            val = 0
            for q in range(self.fe_cell[cell_index].basis_at_quad.shape[1]):
                x = self.fe_cell[cell_index].quad_actual_coordinates[q, 0]
                y = self.fe_cell[cell_index].quad_actual_coordinates[q, 1]
                # print("f_values[q] = ",f_values[q])

                # the Jacobian and the quadrature weights are pre multiplied to the basis functions
                val += (self.fe_cell[cell_index].basis_at_quad[i, q]) * self.fe_cell[
                    cell_index
                ].forcing_function(x, y)
                # print("val = ", val)

            f_integral[i] = val

        self.fe_cell[cell_index].forcing_at_quad = f_integral

        return self.fe_cell[cell_index].forcing_at_quad.copy()

    def get_forcing_function_values_vector(self, cell_index, component) -> np.ndarray:
        """
        This function will return the forcing function values at the quadrature points
        based on the Component of the RHS Needed, for vector valued problems

        :param cell_index: The index of the cell
        :type cell_index: int
        :param component: The component of the RHS needed
        :type component: int
        :return: The forcing function values at the quadrature points
        :rtype: np.ndarray
        :raises ValueError: If cell_index is greater than the number of cells
        """
        if cell_index >= len(self.fe_cell) or cell_index < 0:
            raise ValueError(
                f"cell_index should be less than {self.n_cells} and greater than or equal to 0"
            )

        # get the coordinates
        x = self.fe_cell[cell_index].quad_actual_coordinates[:, 0]
        y = self.fe_cell[cell_index].quad_actual_coordinates[:, 1]

        # compute the forcing function values
        f_values = self.fe_cell[cell_index].forcing_function(x, y)[component]

        # compute the integral
        f_integral = np.sum(self.fe_cell[cell_index].basis_at_quad * f_values, axis=1)

        self.fe_cell[cell_index].forcing_at_quad = f_integral.reshape(-1, 1)

        return self.fe_cell[cell_index].forcing_at_quad.copy()

    def get_sensor_data(self, exact_solution, num_points):
        """
        Obtain sensor data (actual solution) at random points.

        This method is used in the inverse problem to obtain the sensor data at random points within the domain.
        Currently, it only works for problems with an analytical solution.
        Methodologies to obtain sensor data for problems from a file are not implemented yet.
        It is also not implemented for external or complex meshes.

        :param exact_solution: A function that computes the exact solution at a given point.
        :type exact_solution: function
        :param num_points: The number of random points to generate.
        :type num_points: int
        :return: A tuple containing the generated points and the exact solution at those points.
        :rtype: tuple(numpy.ndarray, numpy.ndarray)
        """
        # generate random points within the bounds of the domain
        # get the bounds of the domain
        x_min = np.min(self.mesh.points[:, 0])
        x_max = np.max(self.mesh.points[:, 0])
        y_min = np.min(self.mesh.points[:, 1])
        y_max = np.max(self.mesh.points[:, 1])
        # sample n random points within the bounds of the domain
        # Generate points in the unit square

        num_internal_points = int(num_points * 0.9)

        points = lhs(2, samples=num_internal_points)
        points[:, 0] = x_min + (x_max - x_min) * points[:, 0]
        points[:, 1] = y_min + (y_max - y_min) * points[:, 1]
        # get the exact solution at the points
        exact_sol = exact_solution(points[:, 0], points[:, 1])

        # print the shape of the points and the exact solution
        print(f"[INFO] : Number of sensor points = {points.shape[0]}")
        print(f"[INFO] : Shape of sensor points = {points.shape}")

        # plot the points
        plt.figure(figsize=(6.4, 4.8), dpi=300)
        plt.scatter(points[:, 0], points[:, 1], marker="x", color="r", s=2)
        plt.axis("equal")
        plt.title("Sensor Points")
        plt.tight_layout()
        plt.savefig("sensor_points.png", bbox_inches="tight")

        return points, exact_sol

    def get_sensor_data_external(self, exact_sol, num_points, file_name):
        """
        This method is used to obtain the sensor data from an external file.

        :param exact_sol: The exact solution values.
        :type exact_sol: array-like
        :param num_points: The number of points to sample from the data.
        :type num_points: int
        :param file_name: The path to the file containing the sensor data.
        :type file_name: str

        :return: A tuple containing two arrays:
            - points (ndarray): The sampled points from the data.
            - exact_sol (ndarray): The corresponding exact solution values.
        :rtype: tuple
        """
        # use pandas to read the file
        df = pd.read_csv(file_name)

        x = df.iloc[:, 0].values
        y = df.iloc[:, 1].values
        exact_sol = df.iloc[:, 2].values

        # now sample num_points from the data
        indices = np.random.randint(0, x.shape[0], num_points)

        x = x[indices]
        y = y[indices]
        exact_sol = exact_sol[indices]

        # stack them together
        points = np.stack((x, y), axis=1)

        return points, exact_sol



"""
The file `quad_bilinear.py` defines the Quad Bilinear transformation of the reference element.
The implementation is referenced from the ParMooN project (File: QuadBilineare.C).

Author: Thivin Anandh D

Changelog: 30/Aug/2023 - Initial version

Known issues: Second derivative Calculations are not implemented as of now. 

Dependencies: None specified
"""

import numpy as np
from .fe_transformation_2d import FETransforamtion2D


class QuadBilinear(FETransforamtion2D):
    """
    Defines the Quad Bilinear transformation of the reference element.

    :param co_ordinates: The coordinates of the reference element.
    :type co_ordinates: numpy.ndarray
    """

    def __init__(self, co_ordinates) -> None:
        """
        Constructor for the QuadBilinear class.

        :param co_ordinates: The coordinates of the reference element.
        :type co_ordinates: numpy.ndarray
        """
        self.co_ordinates = co_ordinates
        self.set_cell()
        self.detjk = None  # Jacobian of the transformation

    def set_cell(self):
        """
        Set the cell coordinates, which will be used as intermediate values to calculate the Jacobian and actual values.

        :param None:
        :type None:

        :returns: None
        :rtype: None
        """
        self.x0 = self.co_ordinates[0][0]
        self.x1 = self.co_ordinates[1][0]
        self.x2 = self.co_ordinates[2][0]
        self.x3 = self.co_ordinates[3][0]

        # get the y-coordinates of the cell
        self.y0 = self.co_ordinates[0][1]
        self.y1 = self.co_ordinates[1][1]
        self.y2 = self.co_ordinates[2][1]
        self.y3 = self.co_ordinates[3][1]

        self.xc0 = (self.x0 + self.x1 + self.x2 + self.x3) * 0.25
        self.xc1 = (-self.x0 + self.x1 + self.x2 - self.x3) * 0.25
        self.xc2 = (-self.x0 - self.x1 + self.x2 + self.x3) * 0.25
        self.xc3 = (self.x0 - self.x1 + self.x2 - self.x3) * 0.25

        self.yc0 = (self.y0 + self.y1 + self.y2 + self.y3) * 0.25
        self.yc1 = (-self.y0 + self.y1 + self.y2 - self.y3) * 0.25
        self.yc2 = (-self.y0 - self.y1 + self.y2 + self.y3) * 0.25
        self.yc3 = (self.y0 - self.y1 + self.y2 - self.y3) * 0.25

    def get_original_from_ref(self, xi, eta):
        """
        This method returns the original coordinates from the reference coordinates.

        :param xi: The xi coordinate in the reference element.
        :type xi: float
        :param eta: The eta coordinate in the reference element.
        :type eta: float

        :returns: The original coordinates [x, y] corresponding to the given reference coordinates.
        :rtype: numpy.ndarray
        """
        x = self.xc0 + self.xc1 * xi + self.xc2 * eta + self.xc3 * xi * eta
        y = self.yc0 + self.yc1 * xi + self.yc2 * eta + self.yc3 * xi * eta

        return np.array([x, y], dtype=np.float64)

    def get_jacobian(self, xi, eta):
        """
        This method returns the Jacobian of the transformation.

        :param xi: The xi coordinate in the reference element.
        :type xi: float
        :param eta: The eta coordinate in the reference element.
        :type eta: float

        :returns: The Jacobian of the transformation at the given reference coordinates.
        :rtype: float
        """
        self.detjk = abs(
            (self.xc1 + self.xc3 * eta) * (self.yc2 + self.yc3 * xi)
            - (self.xc2 + self.xc3 * xi) * (self.yc1 + self.yc3 * eta)
        )
        return self.detjk

    def get_orig_from_ref_derivative(self, ref_gradx, ref_grady, xi, eta):
        """
        This method returns the derivatives of the original coordinates with respect to the reference coordinates.

        :param ref_gradx: The gradient of the xi coordinate in the reference element.
        :type ref_gradx: numpy.ndarray
        :param ref_grady: The gradient of the eta coordinate in the reference element.
        :type ref_grady: numpy.ndarray
        :param xi: The xi coordinate in the reference element.
        :type xi: float
        :param eta: The eta coordinate in the reference element.
        :type eta: float

        :returns: The derivatives of the original coordinates [x, y] with respect to the reference coordinates.
        :rtype: numpy.ndarray
        """
        n_test = ref_gradx.shape[0]
        gradx_orig = np.zeros(ref_gradx.shape, dtype=np.float64)
        grady_orig = np.zeros(ref_grady.shape, dtype=np.float64)

        for j in range(n_test):
            Xi = xi
            Eta = eta
            rec_detjk = 1 / (
                (self.xc1 + self.xc3 * Eta) * (self.yc2 + self.yc3 * Xi)
                - (self.xc2 + self.xc3 * Xi) * (self.yc1 + self.yc3 * Eta)
            )
            gradx_orig[j] = (
                (self.yc2 + self.yc3 * Xi) * ref_gradx[j]
                - (self.yc1 + self.yc3 * Eta) * ref_grady[j]
            ) * rec_detjk
            grady_orig[j] = (
                -(self.xc2 + self.xc3 * Xi) * ref_gradx[j]
                + (self.xc1 + self.xc3 * Eta) * ref_grady[j]
            ) * rec_detjk

        return gradx_orig, grady_orig

    def get_orig_from_ref_second_derivative(self, grad_xx_ref, grad_xy_ref, grad_yy_ref, xi, eta):
        """
        This method returns the second derivatives of the original coordinates with respect to the reference coordinates.

        :param grad_xx_ref: The second derivative of the xi coordinate in the reference element.
        :type grad_xx_ref: numpy.ndarray
        :param grad_xy_ref: The mixed second derivative of the xi and eta coordinates in the reference element.
        :type grad_xy_ref: numpy.ndarray
        :param grad_yy_ref: The second derivative of the eta coordinate in the reference element.
        :type grad_yy_ref: numpy.ndarray
        :param xi: The xi coordinate in the reference element.
        :type xi: float
        :param eta: The eta coordinate in the reference element.
        :type eta: float

        :returns: The second derivatives of the original coordinates [xx, xy, yy] with respect to the reference coordinates.
        :rtype: numpy.ndarray
        """
        # print(" Error : Second Derivative not implemented -- Ignore this error, if second derivative is not required ")
        return grad_xx_ref, grad_xy_ref, grad_yy_ref



"""
file: basis_2d_QN_Legendre_Special.py
description: This file contains the class Basis2DQNLegendreSpecial which is used 
             to define the basis functions for a 2D Quad element using a Legendre polynomial.
authors: Thivin Anandh D
changelog: 30/Aug/2023 - Initial version
known_issues: None
"""

import numpy as np

# import the legendre polynomials
from scipy.special import legendre
import matplotlib.pyplot as plt

from .basis_function_2d import BasisFunction2D


class Basis2DQNLegendreSpecial(BasisFunction2D):
    """
    This class defines the basis functions for a 2D Q1 element.

    Attributes:
        num_shape_functions (int): The number of shape functions.

    Methods:
        __init__(self, num_shape_functions: int): Initializes the Basis2DQNLegendreSpecial object.
        test_fcn(self, n_test, x): Calculates the test function values for a given number of tests and input values.
        test_grad_fcn(self, n_test, x): Calculates the gradients of the test functions for a given number of tests and input values.
        test_grad_grad_fcn(self, n_test, x): Calculates the second derivatives of the test functions for a given number of tests and input values.
        value(self, xi, eta): Returns the values of the basis functions at the given (xi, eta) coordinates.
        gradx(self, xi, eta): Returns the x-derivatives of the basis functions at the given (xi, eta) coordinates.
        grady(self, xi, eta): Returns the y-derivatives of the basis functions at the given (xi, eta) coordinates.
        gradxx(self, xi, eta): Returns the xx-derivatives of the basis functions at the given (xi, eta) coordinates.
        gradxy(self, xi, eta): Returns the xy-derivatives of the basis functions at the given (xi, eta) coordinates.
        gradyy(self, xi, eta): Returns the yy-derivatives of the basis functions at the given (xi, eta) coordinates.
    """

    def __init__(self, num_shape_functions: int):
        super().__init__(num_shape_functions)

    def test_fcn(self, n_test, x):
        """
        Calculate the test function values for a given number of tests and input values.

        :param n_test: The number of tests to perform.
        :type n_test: int
        :param x: The input value(s) for the test function.
        :type x: float or array-like
        :return: An array containing the test function values for each test.
        :rtype: numpy.ndarray
        """
        test_total = []
        for n in range(1, n_test + 1):
            obj1 = legendre(n + 1)
            obj2 = legendre(n - 1)
            test = obj1(x) - obj2(x)
            test_total.append(test)
        return np.asarray(test_total)

    def test_grad_fcn(self, n_test, x):
        """
        Calculate the gradient of the test function at a given point.

        :param n_test: The number of test functions to calculate.
        :type n_test: int
        :param x: The point at which to evaluate the gradient.
        :type x: float
        :return: An array containing the gradients of the test functions at the given point.
        :rtype: np.ndarray
        """
        test_total = []
        for n in range(1, n_test + 1):
            obj1 = legendre(n + 1).deriv()
            obj2 = legendre(n - 1).deriv()
            test = obj1(x) - obj2(x)
            test_total.append(test)
        return np.asarray(test_total)

    def test_grad_grad_fcn(self, n_test, x):
        """
        Calculate the gradient of the second derivative of a function using Legendre polynomials.

        :param n_test: The number of test cases to evaluate.
        :type n_test: int
        :param x: The input value at which to evaluate the function.
        :type x: float
        :return: An array containing the results of the test cases.
        :rtype: ndarray
        """
        test_total = []
        for n in range(1, n_test + 1):
            obj1 = legendre(n + 1).deriv(2)
            obj2 = legendre(n - 1).deriv(2)
            test = obj1(x) - obj2(x)

            test_total.append(test)
        return np.asarray(test_total)

    def value(self, xi, eta):
        """
        This method returns the values of the basis functions at the given (xi, eta) coordinates.

        :param xi: The xi coordinate.
        :type xi: numpy.ndarray

        :param eta: The eta coordinate.
        :type eta: numpy.ndarray

        :return: The values of the basis functions at the given coordinates.
        :rtype: numpy.ndarray
        """
        values = np.zeros((self.num_shape_functions, len(xi)))

        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))

        test_function_x = self.test_fcn(num_shape_func_in_1d, xi)
        test_function_y = self.test_fcn(num_shape_func_in_1d, eta)

        # Generate an outer product of the test functions to generate the basis functions
        for i in range(num_shape_func_in_1d):
            values[i * num_shape_func_in_1d : (i + 1) * num_shape_func_in_1d, :] = (
                test_function_x[i, :] * test_function_y
            )

        return values

    def gradx(self, xi, eta):
        """
        This method returns the x-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: The xi coordinate.
        :type xi: numpy.ndarray

        :param eta: The eta coordinate.
        :type eta: numpy.ndarray

        :return: The x-derivatives of the basis functions.
        :rtype: numpy.ndarray
        """
        values = np.zeros((self.num_shape_functions, len(xi)))

        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))

        test_function_grad_x = self.test_grad_fcn(num_shape_func_in_1d, xi)
        test_function_y = self.test_fcn(num_shape_func_in_1d, eta)

        # Generate an outer product of the test functions to generate the basis functions
        for i in range(num_shape_func_in_1d):
            values[i * num_shape_func_in_1d : (i + 1) * num_shape_func_in_1d, :] = (
                test_function_grad_x[i, :] * test_function_y
            )

        return values

    def grady(self, xi, eta):
        """
        This method returns the y-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: The xi coordinates.
        :type xi: numpy.ndarray

        :param eta: The eta coordinates.
        :type eta: numpy.ndarray

        :return: The y-derivatives of the basis functions.
        :rtype: numpy.ndarray
        """
        values = np.zeros((self.num_shape_functions, len(xi)))

        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))

        test_function_x = self.test_fcn(num_shape_func_in_1d, xi)
        test_function_grad_y = self.test_grad_fcn(num_shape_func_in_1d, eta)

        # Generate an outer product of the test functions to generate the basis functions
        for i in range(num_shape_func_in_1d):
            values[i * num_shape_func_in_1d : (i + 1) * num_shape_func_in_1d, :] = (
                test_function_x[i, :] * test_function_grad_y
            )

        return values

    def gradxx(self, xi, eta):
        """
        This method returns the xx-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: The xi coordinate.
        :type xi: numpy.ndarray

        :param eta: The eta coordinate.
        :type eta: numpy.ndarray

        :return: The xx-derivatives of the basis functions.
        :rtype: numpy.ndarray
        """
        values = np.zeros((self.num_shape_functions, len(xi)))

        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))

        test_function_grad_grad_x = self.test_grad_grad_fcn(num_shape_func_in_1d, xi)
        test_function_y = self.test_fcn(num_shape_func_in_1d, eta)

        # Generate an outer product of the test functions to generate the basis functions
        for i in range(num_shape_func_in_1d):
            values[i * num_shape_func_in_1d : (i + 1) * num_shape_func_in_1d, :] = (
                test_function_grad_grad_x[i, :] * test_function_y
            )

        return values

    def gradxy(self, xi, eta):
        """
        This method returns the xy-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: The xi coordinate.
        :type xi: numpy.ndarray

        :param eta: The eta coordinate.
        :type eta: numpy.ndarray

        :return: The xy-derivatives of the basis functions.
        :rtype: numpy.ndarray
        """
        values = np.zeros((self.num_shape_functions, len(xi)))

        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))

        test_function_grad_x = self.test_grad_fcn(num_shape_func_in_1d, xi)
        test_function_grad_y = self.test_grad_fcn(num_shape_func_in_1d, eta)

        # Generate an outer product of the test functions to generate the basis functions
        for i in range(num_shape_func_in_1d):
            values[i * num_shape_func_in_1d : (i + 1) * num_shape_func_in_1d, :] = (
                test_function_grad_x[i, :] * test_function_grad_y
            )

        return values

    def gradyy(self, xi, eta):
        """
        This method returns the yy-derivatives of the basis functions at the given (xi, eta) coordinates.

        :param xi: The xi coordinates.
        :type xi: numpy.ndarray

        :param eta: The eta coordinates.
        :type eta: numpy.ndarray

        :return: The yy-derivatives of the basis functions.
        :rtype: numpy.ndarray
        """
        values = np.zeros((self.num_shape_functions, len(xi)))

        num_shape_func_in_1d = int(np.sqrt(self.num_shape_functions))

        test_function_x = self.test_fcn(num_shape_func_in_1d, xi)
        test_function_grad_grad_y = self.test_grad_grad_fcn(num_shape_func_in_1d, eta)

        # Generate an outer product of the test functions to generate the basis functions
        for i in range(num_shape_func_in_1d):
            values[i * num_shape_func_in_1d : (i + 1) * num_shape_func_in_1d, :] = (
                test_function_x[i, :] * test_function_grad_grad_y
            )

        return values



"""
This module, `geometry.py`, contains the `Geometry` Abstract class which defines functions to read mesh from Gmsh and 
generate internal mesh for 2D and 3D geometries. 

Author: Thivin Anandh D

Date: 03/May/2024
"""

from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import meshio
from pyDOE import lhs

import gmsh

from abc import abstractmethod


class Geometry:
    """
    Abstract class which defines functions to read mesh from Gmsh and internal mesh for 2D problems.

    :param mesh_type: The type of mesh to be used.
    :type mesh_type: str
    :param mesh_generation_method: The method used to generate the mesh.
    :type mesh_generation_method: str
    """

    def __init__(self, mesh_type, mesh_generation_method):
        self.mesh_type = mesh_type
        self.mesh_generation_method = mesh_generation_method

    @abstractmethod
    def read_mesh(
        self,
        mesh_file: str,
        boundary_point_refinement_level: int,
        bd_sampling_method: str,
        refinement_level: int,
    ):
        """
        Abstract method to read mesh from Gmsh.

        :param mesh_file: The path to the mesh file.
        :type mesh_file: str
        :param boundary_point_refinement_level: The refinement level of the boundary points.
        :type boundary_point_refinement_level: int
        :param bd_sampling_method: The method used to sample the boundary points.
        :type bd_sampling_method: str
        :param refinement_level: The refinement level of the mesh.
        :type refinement_level: int
        """

    @abstractmethod
    def generate_vtk_for_test(self):
        """
        Generates a VTK from Mesh file (External) or using gmsh (for Internal).

        :return: None
        """

    @abstractmethod
    def get_test_points(self):
        """
        This function is used to extract the test points from the given mesh

        Parameters:
        None

        Returns:
        test_points (numpy.ndarray): The test points for the given domain
        """



"""
The file `basis_function_2d.py` contains a wrapper class for all the finite element basis functions 
used in the FE2D code. The 2D basis functions have methods to return the value 
of the basis function and its derivatives at the reference point (xi, eta).

Author: Thivin Anandh D

Changelog: 30/Aug/2023 - First version

Known issues: None

Dependencies: None specified
"""

from abc import abstractmethod


class BasisFunction2D:
    """
    This class defines the basis functions for a 2D element.

    :num_shape_functions (int): The number of shape functions.
    :value(xi, eta): Evaluates the basis function at the given xi and eta coordinates.
    """

    def __init__(self, num_shape_functions):
        self.num_shape_functions = num_shape_functions

    @abstractmethod
    def value(self, xi, eta):
        """
        Evaluates the basis function at the given xi and eta coordinates.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The value of the basis function at the given coordinates.
        :rtype: float
        """
        pass

    @abstractmethod
    def gradx(self, xi, eta):
        """
        Computes the partial derivative of the basis function with respect to xi.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The partial derivative of the basis function with respect to xi.
        :rtype: float
        """
        pass

    @abstractmethod
    def grady(self, xi, eta):
        """
        Computes the partial derivative of the basis function with respect to eta.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The partial derivative of the basis function with respect to eta.
        :rtype: float
        """
        pass

    @abstractmethod
    def gradxx(self, xi, eta):
        """
        Computes the second partial derivative of the basis function with respect to xi.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The second partial derivative of the basis function with respect to xi.
        :rtype: float
        """
        pass

    @abstractmethod
    def gradxy(self, xi, eta):
        """
        Computes the mixed partial derivative of the basis function with respect to xi and eta.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The mixed partial derivative of the basis function with respect to xi and eta.
        :rtype: float
        """
        pass

    @abstractmethod
    def gradyy(self, xi, eta):
        """
        Computes the second partial derivative of the basis function with respect to eta.

        :param float xi: The xi coordinate.
        :param float eta: The eta coordinate.
        :return: The second partial derivative of the basis function with respect to eta.
        :rtype: float
        """
        pass


# ---------------- Legendre -------------------------- #
from .basis_2d_QN_Legendre import *  # Normal Legendre from Jacobi -> J(n) = J(n-1) - J(n+1)
from .basis_2d_QN_Legendre_Special import *  # L(n) = L(n-1) - L(n+1)

# ---------------- Jacobi -------------------------- #
from .basis_2d_QN_Jacobi import *  # Normal Jacobi

# ---------------- Chebyshev -------------------------- #
from .basis_2d_QN_Chebyshev_2 import *  # Normal Chebyshev



# This class provides an interface for the transformation methods used in the
# 3D finite element analysis. The transformation is essential for mapping
# element geometry in the reference domain to the actual physical domain.
# The primary functionalities encapsulated within this class include:
#   1. set_cell() - To set the physical coordinates of the cell. These coordinates
#                   are essential for subsequent computations involving Jacobians and transformations.
#   2. get_original_from_ref(xi, eta, zeta) - Given reference coordinates (xi, eta, zeta), this
#                   method returns the corresponding coordinates in the physical domain.
#   3. get_jacobian(xi, eta, zeta) - For a given point in the reference domain, represented
#                   by (xi, eta, zeta), this method calculates and returns the Jacobian of
#                   the transformation, which provides information about the local stretching,
#                   rotation, and skewing of the element.
#
# Further implementations of this class for specific element types (like quad and triangular elements)
# can incorporate more detailed and element-specific transformation techniques.
#
# Author: Thivin Anandh D
# Date:   03-May-2024
# History: First version -  03-May-2024 - Thivin Anandh

from abc import abstractmethod


class FETransforamtion3D:  # pragma: no cover
    """
    This class represents a 2D finite element transformation.
    """

    def __init__(self):
        """
        Constructor for the FETransformation3D class.
        """
        pass

    @abstractmethod
    def set_cell(self):
        """
        Set the cell co-ordinates, which will be used to calculate the Jacobian and actual values.

        :return: None
        """

    @abstractmethod
    def get_original_from_ref(self, xi, eta, zeta):
        """
        This method returns the original coordinates from the reference coordinates.

        :param xi: The xi coordinate in the reference space.
        :type xi: float
        :param eta: The eta coordinate in the reference space.
        :type eta: float
        :param zeta: The zeta coordinate in the reference space.
        :type zeta: float
        :return: The original coordinates (x, y) corresponding to the given reference coordinates.
        :rtype: tuple
        """

    @abstractmethod
    def get_original_from_ref(self, xi, eta, zeta):
        """
        This method returns the original co-ordinates from the reference co-ordinates.

        :param xi: The xi value of the reference co-ordinates.
        :type xi: float
        :param eta: The eta value of the reference co-ordinates.
        :type eta: float
        :param zeta: The zeta value of the reference co-ordinates.
        :type zeta: float
        :return: The original co-ordinates corresponding to the given reference co-ordinates.
        :rtype: tuple
        """

    @abstractmethod
    def get_jacobian(self, xi, eta, zeta):
        """
        This method returns the Jacobian of the transformation.

        :param xi: The xi coordinate.
        :type xi: float
        :param eta: The eta coordinate.
        :type eta: float
        :param zeta: The zeta coordinate.
        :type zeta: float
        :return: The Jacobian matrix.
        :rtype: numpy.ndarray
        """



"""
This file `compute_utils.py` contains the utility functions for 
computing the errors between the exact and predicted solutions.

Author: Thivin Anandh D

Date: 02/Nov/2023

Changelog: 02/Nov/2023 - file created; and added functions to compute L1, L2, L_inf errors

Known issues: None
"""

# Importing the required libraries
import numpy as np


def compute_l2_error(u_exact, u_approx):
    """
    This function will compute the L2 error between the exact solution and the approximate solution.

    :param u_exact: numpy array containing the exact solution
    :type u_exact: numpy.ndarray
    :param u_approx: numpy array containing the approximate solution
    :type u_approx: numpy.ndarray

    :return: L2 error between the exact and approximate solutions
    :rtype: float
    """

    # The L2 error is defined as:

    #   ..math::
    #        \\sqrt{\\frac{1}{N} \\sum_{i=1}^{N} (u_{exact} - u_{approx})^2}

    # Flatten the arrays
    u_exact = u_exact.flatten()
    u_approx = u_approx.flatten()

    # compute the L2 error
    l2_error = np.sqrt(np.mean(np.square(u_exact - u_approx)))
    return l2_error


def compute_l1_error(u_exact, u_approx):
    """
    This function will compute the L1 error between the exact solution and the approximate solution.

    :param u_exact: numpy array containing the exact solution
    :type u_exact: numpy.ndarray
    :param u_approx: numpy array containing the approximate solution
    :type u_approx: numpy.ndarray

    :return: L1 error between the exact and approximate solutions
    :rtype: float
    """

    # The L1 error is defined as:

    #    ..math::
    #        \\frac{1}{N} \\sum_{i=1}^{N} |u_{exact} - u_{approx}|

    # Flatten the arrays
    u_exact = u_exact.flatten()
    u_approx = u_approx.flatten()
    # compute the L1 error
    l1_error = np.mean(np.abs(u_exact - u_approx))
    return l1_error


def compute_linf_error(u_exact, u_approx):
    """
    This function will compute the L_inf error between the exact solution and the approximate solution.

    :param u_exact: numpy array containing the exact solution
    :type u_exact: numpy.ndarray
    :param u_approx: numpy array containing the approximate solution
    :type u_approx: numpy.ndarray

    :return: L_inf error between the exact and approximate solutions
    :rtype: float
    """

    # The L_inf error is defined as

    #    ..math::
    #        \\max_{i=1}^{N} |u_{exact} - u_{approx}|

    # flatten the arrays
    u_exact = u_exact.flatten()
    u_approx = u_approx.flatten()

    # compute the L_inf error
    linf_error = np.max(np.abs(u_exact - u_approx))
    return linf_error


def compute_l2_error_relative(u_exact, u_approx):
    """
    This function will compute the relative L2 error between the exact solution and the approximate solution.

    :param u_exact: numpy array containing the exact solution
    :type u_exact: numpy.ndarray
    :param u_approx: numpy array containing the approximate solution
    :type u_approx: numpy.ndarray

    :return: relative L2 error between the exact and approximate solutions
    :rtype: float
    """

    # The relative L2 error is defined as:

    #    ..math::
    #        \\frac{\\sqrt{\\frac{1}{N} \\sum_{i=1}^{N} (u_{exact} - u_{approx})^2}}{\\sqrt{\\frac{1}{N} \\sum_{i=1}^{N} u_{exact}^2}}

    # flatten the arrays
    u_exact = u_exact.flatten()
    u_approx = u_approx.flatten()

    # compute the L2 error
    l2_error = compute_l2_error(u_exact, u_approx)
    # compute the relative L2 error
    l2_error_relative = l2_error / np.sqrt(np.mean(np.square(u_exact)))
    return l2_error_relative


def compute_linf_error_relative(u_exact, u_approx):
    """
    This function will compute the relative L_inf error between the exact solution and the approximate solution.

    :param u_exact: numpy array containing the exact solution
    :type u_exact: numpy.ndarray
    :param u_approx: numpy array containing the approximate solution
    :type u_approx: numpy.ndarray

    :return: relative L_inf error between the exact and approximate solutions
    :rtype: float
    """

    # The relative L_inf error is defined as:

    #    ..math::
    #        \\frac{\\max_{i=1}^{N} |u_{exact} - u_{approx}|}{\\max_{i=1}^{N} |u_{exact}|}

    # flatten the arrays
    u_exact = u_exact.flatten()
    u_approx = u_approx.flatten()

    # compute the L_inf error
    linf_error = compute_linf_error(u_exact, u_approx)
    # compute the relative L_inf error
    linf_error_relative = linf_error / np.max(np.abs(u_exact))
    return linf_error_relative


def compute_l1_error_relative(u_exact, u_approx):
    """
    This function will compute the relative L1 error between the exact solution and the approximate solution.

    :param u_exact: numpy array containing the exact solution
    :type u_exact: numpy.ndarray
    :param u_approx: numpy array containing the approximate solution
    :type u_approx: numpy.ndarray

    :return: relative L1 error between the exact and approximate solutions
    :rtype: float
    """

    # The relative L1 error is defined as:

    #    ..math::
    #        \\frac{\\frac{1}{N} \\sum_{i=1}^{N} |u_{exact} - u_{approx}|}{\\frac{1}{N} \\sum_{i=1}^{N} |u_{exact}|}

    # flatten the arrays
    u_exact = u_exact.flatten()
    u_approx = u_approx.flatten()

    # compute the L2 error
    l1_error = compute_l1_error(u_exact, u_approx)
    # compute the relative l1 error
    l1_error_relative = l1_error / np.mean(np.abs(u_exact))
    return l1_error_relative


def compute_errors_combined(u_exact, u_approx):
    """
    This function will compute the L1, L2 and L_inf absolute and relative errors.

    :param u_exact: numpy array containing the exact solution
    :type u_exact: numpy.ndarray
    :param u_approx: numpy array containing the approximate solution
    :type u_approx: numpy.ndarray

    :return: L1, L2 and L_inf absolute and relative errors
    :rtype: tuple

    """
    # flatten the arrays
    u_exact = u_exact.flatten()
    u_approx = u_approx.flatten()

    # compute the L2 error
    l2_error = compute_l2_error(u_exact, u_approx)
    # compute the L_inf error
    linf_error = compute_linf_error(u_exact, u_approx)
    # compute the relative L2 error
    l2_error_relative = compute_l2_error_relative(u_exact, u_approx)
    # compute the relative L_inf error
    linf_error_relative = compute_linf_error_relative(u_exact, u_approx)

    # compute L1 Error
    l1_error = compute_l1_error(u_exact, u_approx)

    # compute the relative L1 error
    l1_error_relative = compute_l1_error_relative(u_exact, u_approx)

    return (
        l2_error,
        linf_error,
        l2_error_relative,
        linf_error_relative,
        l1_error,
        l1_error_relative,
    )



"""
The file `model.py` hosts the Neural Network (NN) model and the training loop for variational Physics-Informed Neural Networks (PINNs).
The focus is on the model architecture and the training loop, and not on the loss functions.

Author: Thivin Anandh D

Changelog: 22/Sep/2023 - Initial implementation with basic model architecture and training loop

Known issues: None

Dependencies: None specified
"""

import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras import initializers
import copy


# Custom Model
class DenseModel(tf.keras.Model):
    """
    Defines the Dense Model for the Neural Network for solving Variational PINNs.

    :param layer_dims: List of dimensions of the dense layers.
    :type layer_dims: list
    :param learning_rate_dict: The dictionary containing the learning rate parameters.
    :type learning_rate_dict: dict
    :param params_dict: The dictionary containing the parameters.
    :type params_dict: dict
    :param loss_function: The loss function for the PDE.
    :type loss_function: function
    :param input_tensors_list: The list containing the input tensors.
    :type input_tensors_list: list
    :param orig_factor_matrices: The list containing the original factor matrices.
    :type orig_factor_matrices: list
    :param force_function_list: The force function matrix.
    :type force_function_list: tf.Tensor
    :param tensor_dtype: The tensorflow dtype to be used for all the tensors.
    :type tensor_dtype: tf.DType
    :param use_attention: Flag to use attention layer after input, defaults to False.
    :type use_attention: bool, optional
    :param activation: The activation function to be used for the dense layers, defaults to "tanh".
    :type activation: str, optional
    :param hessian: Flag to use hessian loss, defaults to False.
    :type hessian: bool, optional
    """

    def __init__(
        self,
        layer_dims,
        learning_rate_dict,
        params_dict,
        loss_function,
        input_tensors_list,
        orig_factor_matrices,
        force_function_list,
        tensor_dtype,
        use_attention=False,
        activation="tanh",
        hessian=False,
    ):
        super(DenseModel, self).__init__()
        self.layer_dims = layer_dims
        self.use_attention = use_attention
        self.activation = activation
        self.layer_list = []
        self.loss_function = loss_function
        self.hessian = hessian

        self.tensor_dtype = tensor_dtype

        # if dtype is not a valid tensorflow dtype, raise an error
        if not isinstance(self.tensor_dtype, tf.DType):
            raise TypeError("The given dtype is not a valid tensorflow dtype")

        self.orig_factor_matrices = orig_factor_matrices
        self.shape_function_mat_list = copy.deepcopy(orig_factor_matrices[0])
        self.shape_function_grad_x_factor_mat_list = copy.deepcopy(orig_factor_matrices[1])
        self.shape_function_grad_y_factor_mat_list = copy.deepcopy(orig_factor_matrices[2])

        self.force_function_list = force_function_list

        self.input_tensors_list = input_tensors_list
        self.input_tensor = copy.deepcopy(input_tensors_list[0])
        self.dirichlet_input = copy.deepcopy(input_tensors_list[1])
        self.dirichlet_actual = copy.deepcopy(input_tensors_list[2])

        self.params_dict = params_dict

        self.pre_multiplier_val = self.shape_function_mat_list
        self.pre_multiplier_grad_x = self.shape_function_grad_x_factor_mat_list
        self.pre_multiplier_grad_y = self.shape_function_grad_y_factor_mat_list

        self.force_matrix = self.force_function_list

        print(f"{'-'*74}")
        print(f"| {'PARAMETER':<25} | {'SHAPE':<25} |")
        print(f"{'-'*74}")
        print(
            f"| {'input_tensor':<25} | {str(self.input_tensor.shape):<25} | {self.input_tensor.dtype}"
        )
        print(
            f"| {'force_matrix':<25} | {str(self.force_matrix.shape):<25} | {self.force_matrix.dtype}"
        )
        print(
            f"| {'pre_multiplier_grad_x':<25} | {str(self.pre_multiplier_grad_x.shape):<25} | {self.pre_multiplier_grad_x.dtype}"
        )
        print(
            f"| {'pre_multiplier_grad_y':<25} | {str(self.pre_multiplier_grad_y.shape):<25} | {self.pre_multiplier_grad_y.dtype}"
        )
        print(
            f"| {'pre_multiplier_val':<25} | {str(self.pre_multiplier_val.shape):<25} | {self.pre_multiplier_val.dtype}"
        )
        print(
            f"| {'dirichlet_input':<25} | {str(self.dirichlet_input.shape):<25} | {self.dirichlet_input.dtype}"
        )
        print(
            f"| {'dirichlet_actual':<25} | {str(self.dirichlet_actual.shape):<25} | {self.dirichlet_actual.dtype}"
        )
        print(f"{'-'*74}")

        self.n_cells = params_dict["n_cells"]

        ## ----------------------------------------------------------------- ##
        ## ---------- LEARNING RATE AND OPTIMISER FOR THE MODEL ------------ ##
        ## ----------------------------------------------------------------- ##

        # parse the learning rate dictionary
        self.learning_rate_dict = learning_rate_dict
        initial_learning_rate = learning_rate_dict["initial_learning_rate"]
        use_lr_scheduler = learning_rate_dict["use_lr_scheduler"]
        decay_steps = learning_rate_dict["decay_steps"]
        decay_rate = learning_rate_dict["decay_rate"]
        # staircase = learning_rate_dict["staircase"]

        if use_lr_scheduler:
            learning_rate_fn = tf.keras.optimizers.schedules.ExponentialDecay(
                initial_learning_rate, decay_steps, decay_rate, staircase=True
            )
        else:
            learning_rate_fn = initial_learning_rate

        self.optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate_fn)

        ## ----------------------------------------------------------------- ##
        ## --------------------- MODEL ARCHITECTURE ------------------------ ##
        ## ----------------------------------------------------------------- ##

        # Build dense layers based on the input list
        for dim in range(len(self.layer_dims) - 2):
            self.layer_list.append(
                layers.Dense(
                    self.layer_dims[dim + 1],
                    activation=self.activation,
                    kernel_initializer="glorot_uniform",
                    dtype=self.tensor_dtype,
                    bias_initializer="zeros",
                )
            )

        # Add a output layer with no activation
        self.layer_list.append(
            layers.Dense(
                self.layer_dims[-1],
                activation=None,
                kernel_initializer="glorot_uniform",
                dtype=self.tensor_dtype,
                bias_initializer="zeros",
            )
        )

        # Add attention layer if required
        if self.use_attention:
            self.attention_layer = layers.Attention()

        # Compile the model
        self.compile(optimizer=self.optimizer)
        self.build(input_shape=(None, self.layer_dims[0]))

        # print the summary of the model
        self.summary()

    # def build(self, input_shape):
    #     super(DenseModel, self).build(input_shape)

    def call(self, inputs):
        """
        The call method for the model.

        :param inputs: The input tensor for the model.
        :type inputs: tf.Tensor
        :return: The output tensor of the model.
        :rtype: tf.Tensor
        """
        x = inputs

        # Apply attention layer after input if flag is True
        if self.use_attention:
            x = self.attention_layer([x, x])

        # Loop through the dense layers
        for layer in self.layer_list:
            x = layer(x)

        return x

    def get_config(self):
        """
        Get the configuration of the model.

        Returns:
            dict: The configuration of the model.
        """
        # Get the base configuration
        base_config = super().get_config()

        # Add the non-serializable arguments to the configuration
        base_config.update(
            {
                "learning_rate_dict": self.learning_rate_dict,
                "loss_function": self.loss_function,
                "input_tensors_list": self.input_tensors_list,
                "orig_factor_matrices": self.orig_factor_matrices,
                "force_function_list": self.force_function_list,
                "params_dict": self.params_dict,
                "use_attention": self.use_attention,
                "activation": self.activation,
                "hessian": self.hessian,
                "layer_dims": self.layer_dims,
                "tensor_dtype": self.tensor_dtype,
            }
        )

        return base_config

    @tf.function
    def train_step(self, beta=10, bilinear_params_dict=None):  # pragma: no cover
        """
        The train step method for the model.

        :param beta: The beta parameter for the training step, defaults to 10.
        :type beta: int, optional
        :param bilinear_params_dict: The dictionary containing the bilinear parameters, defaults to None.
        :type bilinear_params_dict: dict, optional
        :return: The output of the training step.
        :rtype: varies based on implementation
        """

        with tf.GradientTape(persistent=True) as tape:
            # Predict the values for dirichlet boundary conditions
            predicted_values_dirichlet = self(self.dirichlet_input)

            # initialize total loss as a tensor with shape (1,) and value 0.0
            total_pde_loss = 0.0

            with tf.GradientTape(persistent=True) as tape1:
                # tape gradient
                tape1.watch(self.input_tensor)
                # Compute the predicted values from the model
                predicted_values = self(self.input_tensor)

            # compute the gradients of the predicted values wrt the input which is (x, y)
            gradients = tape1.gradient(predicted_values, self.input_tensor)

            # Split the gradients into x and y components and reshape them to (-1, 1)
            # the reshaping is done for the tensorial operations purposes (refer Notebook)
            pred_grad_x = tf.reshape(
                gradients[:, 0], [self.n_cells, self.pre_multiplier_grad_x.shape[-1]]
            )  # shape : (N_cells , N_quadrature_points)
            pred_grad_y = tf.reshape(
                gradients[:, 1], [self.n_cells, self.pre_multiplier_grad_y.shape[-1]]
            )  # shape : (N_cells , N_quadrature_points)

            pred_val = tf.reshape(
                predicted_values, [self.n_cells, self.pre_multiplier_val.shape[-1]]
            )  # shape : (N_cells , N_quadrature_points)

            cells_residual = self.loss_function(
                test_shape_val_mat=self.pre_multiplier_val,
                test_grad_x_mat=self.pre_multiplier_grad_x,
                test_grad_y_mat=self.pre_multiplier_grad_y,
                pred_nn=pred_val,
                pred_grad_x_nn=pred_grad_x,
                pred_grad_y_nn=pred_grad_y,
                forcing_function=self.force_matrix,
                bilinear_params=bilinear_params_dict,
            )

            residual = tf.reduce_sum(cells_residual)

            # Compute the total loss for the PDE
            total_pde_loss = total_pde_loss + residual

            # print shapes of the predicted values and the actual values
            boundary_loss = tf.reduce_mean(
                tf.square(predicted_values_dirichlet - self.dirichlet_actual), axis=0
            )

            # Compute Total Loss
            total_loss = total_pde_loss + beta * boundary_loss

        trainable_vars = self.trainable_variables
        self.gradients = tape.gradient(total_loss, trainable_vars)
        self.optimizer.apply_gradients(zip(self.gradients, trainable_vars))

        return {"loss_pde": total_pde_loss, "loss_dirichlet": boundary_loss, "loss": total_loss}



from setuptools import setup, find_packages
import os

# Read the requirements from the requirements.txt file
with open("requirements.txt") as f:
    requirements = f.read().splitlines()


# Moved all the metadata to the pyproject.toml file
setup(
    name="fastvpinns",
    # version="1.0.5",
    # author="Thivin Anandh, Divij Ghose, Sashikumaar Ganesan",
    # author_email="thivinanandh@gmail.com",
    # description="An Efficient tensor-based implementation of hp-Variational Physics-informed Neural Networks (VPINNs) for solving PDEs",
    packages=find_packages(),
    # long_description=read_file('README.md'),
    # home_page="https://cmgcds.github.io/fastvpinns",
    # license="MIT",
    install_requires=requirements,
    # keywords=['variational PINNs', 'VPINNs', 'PINNs', 'Physics-informed neural networks', 'Deep learning', 'Machine learning']
)



"""
The file `quadratureformulas_quad2d.py` defines the Quadrature Formulas for the 2D Quadrilateral elements.
It supports both Gauss-Legendre and Gauss-Jacobi quadrature types.
The quadrature points and weights are calculated based on the specified quadrature order and type.

Author: Thivin Anandh D

Changelog: Not specified

Known issues: None

Dependencies: numpy, scipy
"""

import numpy as np
from scipy.special import roots_legendre, roots_jacobi, jacobi, gamma
from scipy.special import legendre
from scipy.special import eval_legendre, legendre

from .quadratureformulas import Quadratureformulas


class Quadratureformulas_Quad2D(Quadratureformulas):
    """
    Defines the Quadrature Formulas for the 2D Quadrilateral elements.

    :param quad_order: The order of the quadrature.
    :type quad_order: int
    :param quad_type: The type of the quadrature.
    :type quad_type: str
    """

    def __init__(self, quad_order: int, quad_type: str):
        """
        Constructor for the Quadratureformulas_Quad2D class.

        :param quad_order: The order of the quadrature.
        :type quad_order: int
        :param quad_type: The type of the quadrature.
        :type quad_type: str
        """
        # initialize the super class
        super().__init__(
            quad_order=quad_order, quad_type=quad_type, num_quad_points=quad_order * quad_order
        )

        # Calculate the Gauss-Legendre quadrature points and weights for 1D
        # nodes_1d, weights_1d = roots_jacobi(self.quad_order, 1, 1)

        quad_type = self.quad_type

        if quad_type == "gauss-legendre":
            # Commented out by THIVIN -  to Just use legendre quadrature points as it is
            # if quad_order == 2:
            #     nodes_1d = np.array([-1, 1])
            #     weights_1d = np.array([1, 1])
            # else:
            nodes_1d, weights_1d = np.polynomial.legendre.leggauss(quad_order)  # Interior points
            # nodes_1d = np.concatenate(([-1, 1], nodes_1d))
            # weights_1d = np.concatenate(([1, 1], weights_1d))

            # Generate the tensor outer product of the nodes
            xi_quad, eta_quad = np.meshgrid(nodes_1d, nodes_1d)
            xi_quad = xi_quad.flatten()
            eta_quad = eta_quad.flatten()

            # Multiply the weights accordingly for 2D
            quad_weights = (weights_1d[:, np.newaxis] * weights_1d).flatten()

            # Assign the values
            self.xi_quad = xi_quad
            self.eta_quad = eta_quad
            self.quad_weights = quad_weights

        elif quad_type == "gauss-jacobi":

            def GaussJacobiWeights(Q: int, a, b):
                [X, W] = roots_jacobi(Q, a, b)
                return [X, W]

            def jacobi_wrapper(n, a, b, x):

                x = np.array(x, dtype=np.float64)

                return jacobi(n, a, b)(x)

            # Weight coefficients
            def GaussLobattoJacobiWeights(Q: int, a, b):
                W = []
                X = roots_jacobi(Q - 2, a + 1, b + 1)[0]
                if a == 0 and b == 0:
                    W = 2 / ((Q - 1) * (Q) * (jacobi_wrapper(Q - 1, 0, 0, X) ** 2))
                    Wl = 2 / ((Q - 1) * (Q) * (jacobi_wrapper(Q - 1, 0, 0, -1) ** 2))
                    Wr = 2 / ((Q - 1) * (Q) * (jacobi_wrapper(Q - 1, 0, 0, 1) ** 2))
                else:
                    W = (
                        2 ** (a + b + 1)
                        * gamma(a + Q)
                        * gamma(b + Q)
                        / (
                            (Q - 1)
                            * gamma(Q)
                            * gamma(a + b + Q + 1)
                            * (jacobi_wrapper(Q - 1, a, b, X) ** 2)
                        )
                    )
                    Wl = (
                        (b + 1)
                        * 2 ** (a + b + 1)
                        * gamma(a + Q)
                        * gamma(b + Q)
                        / (
                            (Q - 1)
                            * gamma(Q)
                            * gamma(a + b + Q + 1)
                            * (jacobi_wrapper(Q - 1, a, b, -1) ** 2)
                        )
                    )
                    Wr = (
                        (a + 1)
                        * 2 ** (a + b + 1)
                        * gamma(a + Q)
                        * gamma(b + Q)
                        / (
                            (Q - 1)
                            * gamma(Q)
                            * gamma(a + b + Q + 1)
                            * (jacobi_wrapper(Q - 1, a, b, 1) ** 2)
                        )
                    )
                W = np.append(W, Wr)
                W = np.append(Wl, W)
                X = np.append(X, 1)
                X = np.append(-1, X)
                return [X, W]

            # get quadrature points and weights in 1D
            x, w = GaussLobattoJacobiWeights(self.quad_order, 0, 0)

            # Generate the tensor outer product of the nodes
            xi_quad, eta_quad = np.meshgrid(x, x)
            xi_quad = xi_quad.flatten()
            eta_quad = eta_quad.flatten()

            # Multiply the weights accordingly for 2D
            quad_weights = (w[:, np.newaxis] * w).flatten()

            # Assign the values
            self.xi_quad = xi_quad
            self.eta_quad = eta_quad
            self.quad_weights = quad_weights

        else:
            print("Supported quadrature types are: gauss-legendre, gauss-jacobi")
            print(
                f"Invalid quadrature type {quad_type} in {self.__class__.__name__} from {__name__}."
            )
            raise ValueError("Quadrature type not supported.")

    def get_quad_values(self):
        """
        Returns the quadrature weights, xi and eta values.

        :return: A tuple containing the quadrature weights, xi values, and eta values.
        :rtype: tuple
        """
        return self.quad_weights, self.xi_quad, self.eta_quad

    def get_num_quad_points(self):
        """
        Returns the number of quadrature points.

        :return: The number of quadrature points.
        :rtype: int
        """
        return self.num_quad_points
